"""
Stores the functions required for adjusting time factor in the neighbouring
waypoints in Synfig
"""

import sys
sys.path.append("../")


def time_adjust(lottie, animated):
    """
    Adjusts the tangents between neighbouring waypoints depending upon the time
    factor between previous waypoints or next waypoints

    Args:
        lottie   (dict)                : Holds bezier curve in Lottie format
        animated (lxml.etree._Element) : Synfig format animation

    Returns:
        (None)
    """
    timeadjust = 0.5
    for i in range(len(animated) - 1):
        if i == 0:
            continue
        time_span_cur = lottie["k"][i+1]["t"] - lottie["k"][i]["t"]
        time_span_prev = lottie["k"][i]["t"] - lottie["k"][i-1]["t"]
        cur_get_after = animated[i].attrib["after"]
        next_get_before = animated[i+1].attrib["before"]

        # prev              iter
        # ANY/CONSTANT ---- ANY/ANY
        # ANY/ANY      ---- CONSTANT/ANY
        if cur_get_after == "constant" or next_get_before == "constant":
            continue

        if animated.attrib["type"] == "real":
            if cur_get_after != "linear":
                lottie["k"][i]["o"]["x"][0] *= (time_span_cur * (timeadjust + 1)) /\
                        (time_span_cur * timeadjust + time_span_prev)
                lottie["k"][i]["o"]["y"][0] *= (time_span_cur * (timeadjust + 1)) /\
                        (time_span_cur * timeadjust + time_span_prev)
            if next_get_before != "linear":
                if i + 2 <= len(animated) - 1:
                    time_span_next = lottie["k"][i+2]["t"] - lottie["k"][i+1]["t"]
                    lottie["k"][i]["i"]["x"][0] *= (time_span_cur * (timeadjust + 1)) /\
                            (time_span_cur * timeadjust + time_span_next)

        elif animated.attrib["type"] == "vector":

            # prev    --- iter        --- next
            # ANY/ANY --- ANY/!LINEAR --- ANY/ANY
            if cur_get_after != "linear":
                for dim in range(len(lottie["k"][i]["to"])):
                    lottie["k"][i]["to"][dim] = lottie["k"][i]["to"][dim] *\
                    (time_span_cur * (timeadjust + 1)) /\
                    (time_span_cur * timeadjust + time_span_prev)

            # iter    --- next        --- after_next
            # ANY/ANY --- !LINEAR/ANY --- ANY/ANY
            if next_get_before != "linear":
                for dim in range(len(lottie["k"][i]["to"])):
                    if i + 2 <= len(animated) - 1:
                        time_span_next = lottie["k"][i+2]["t"] - lottie["k"][i+1]["t"]
                        lottie["k"][i]["ti"][dim] = lottie["k"][i]["ti"][dim] *\
                        (time_span_cur * (timeadjust + 1)) /\
                        (time_span_cur * timeadjust + time_span_next)



# pylint: disable=line-too-long
"""
Will store all functions needed to generate the rectangle layer in lottie
"""

import sys
import copy
from lxml import etree
import settings
from common.Count import Count
from common.misc import is_animated, get_frame, get_vector, set_vector
from properties.value import gen_properties_value
from properties.multiDimensionalKeyframed import gen_properties_multi_dimensional_keyframed
from properties.valueKeyframed import gen_value_Keyframed
from synfig.animation import gen_dummy_waypoint, insert_waypoint_at_frame, to_Synfig_axis, get_vector_at_frame, get_animated_time_list, copy_tcb_average, copy_tcb
sys.path.append("..")


def get_child_value(is_animate, child, what_type):
    """
    Depending upon the is_animate type, value of the position or other
    parameters are extracted by this function

    Args:
        is_animate (int)                : Decides whether a parameter is animated
        child      (lxml.etree._Element): Holds the waypoint values
        what_type  (str)                : Decides the type of waypoint

    Returns:
        (float, float)    :   If the type of waypoint is "position"
        (float)           :   If the type of waypoint is "value"
    """
    if what_type == "position":
        if is_animate == settings.NOT_ANIMATED:
            x_val, y_val = float(child[0][0].text), float(child[0][1].text)
        elif is_animate == settings.SINGLE_WAYPOINT:
            x_val, y_val = float(child[0][0][0][0].text), float(child[0][0][0][1].text)
        x_val *= settings.PIX_PER_UNIT
        y_val *= settings.PIX_PER_UNIT
        return x_val, y_val
    elif what_type == "value":
        if is_animate == settings.NOT_ANIMATED:
            val = float(child[0].attrib["value"])
        elif is_animate == settings.SINGLE_WAYPOINT:
            val = float(child[0][0][0].attrib["value"])
        return val


def gen_shapes_rectangle(lottie, layer, idx):
    """
    Generates the dictionary corresponding to shapes/rect.json

    Args:
        lottie (dict)               : The lottie generated rectangle layer will be stored in it
        layer  (lxml.etree._Element): Synfig format rectangle layer
        idx    (int)                : Stores the index of the rectangle layer

    Returns:
        (None)
    """
    index = Count()
    lottie["ty"] = "rc"     # Type: rectangle
    lottie["p"] = {}        # Position of rectangle
    lottie["d"] = settings.DEFAULT_DIRECTION
    lottie["s"] = {}        # Size of rectangle
    lottie["ix"] = idx      # setting the index
    lottie["r"] = {}        # Rounded corners of rectangle
    points = {}
    bevel_found = False
    expand_found = False    # For filled rectangle layers

    for child in layer:
        if child.tag == "param":
            if child.attrib["name"] == "point1":
                points["1"] = child # Store address of child here

            elif child.attrib["name"] == "point2":
                points["2"] = child # Store address of child here

            elif child.attrib["name"] == "expand":
                expand_found = True
                param_expand = child

            elif child.attrib["name"] == "bevel":
                bevel_found = True
                is_animate = is_animated(child[0])
                if is_animate == settings.ANIMATED:
                    gen_value_Keyframed(lottie["r"], child[0], index.inc())
                else:
                    bevel = get_child_value(is_animate, child, "value")
                    bevel *= settings.PIX_PER_UNIT
                    gen_properties_value(lottie["r"],
                                         bevel,
                                         index.inc(),
                                         settings.DEFAULT_ANIMATED,
                                         settings.NO_INFO)
    if not bevel_found:     # For rectangle layer in stable version 1.2.2
        bevel = 0
        gen_properties_value(lottie["r"], bevel, index.inc(),
                             settings.DEFAULT_ANIMATED, settings.NO_INFO)

    if not expand_found:    # Means filled rectangle layer, gen expand param
        st = "<param name='expand'><real value='0.0'/></param>"
        param_expand = etree.fromstring(st)

    # If expand parameter is not animated
    param_expand = gen_dummy_waypoint(param_expand, "param", "real")

    # expand parameter needs to be positive: required by Synfig
    make_positive_valued(param_expand)

    # If p1 not animated
    points["1"] = gen_dummy_waypoint(points["1"], "param", "vector")

    # If p2 not animated
    points["2"] = gen_dummy_waypoint(points["2"], "param", "vector")

    both_points_animated(points["1"], points["2"], param_expand, lottie, index)


def make_positive_valued(param):
    """
    User might mistakenly provide negative values to expand parameter; This is
    taken care of in this function
    """
    animated = param[0]
    for waypoint in animated:
        val = float(waypoint[0].attrib["value"])
        val = abs(val)
        waypoint[0].attrib["value"] = str(val)


def both_points_animated(animated_1, animated_2, param_expand, lottie, index):
    """
    This function generates the lottie dictionary for position and size property
    of lottie(point1 and point2 are used from Synfig), when both point1 and
    point2 are animated

    Args:
        animated_1      (lxml.etree._Element): Holds the parameter `point1`'s animation in Synfig xml format
        animated_2      (lxml.etree._Element): Holds the parameter `point2`'s animation in Synfig xml format
        param_expand    (lxml.etree._Element): Holds the parameter `expand`'s animation in Synfig xml format
        lottie          (dict)               : Lottie format rectangle layer will be store in this
        index           (int)                : Stores the index of parameters in rectangle layer

    Returns:
        (None)
    """
    animated_1, animated_2 = animated_1[0], animated_2[0]
    orig_path_1, orig_path_2 = {}, {}
    expand_path = {}
    gen_value_Keyframed(expand_path, param_expand[0], 0)
    gen_properties_multi_dimensional_keyframed(orig_path_1, animated_1, 0)
    gen_properties_multi_dimensional_keyframed(orig_path_2, animated_2, 0)

    #################### SECTION 1 ###########################
    # Insert waypoints in the point1 and point2 parameter at the place where
    # expand parameter is animated
    time_list = set()
    get_animated_time_list(param_expand, time_list)
    for frame in time_list:
        insert_waypoint_at_frame(animated_1, orig_path_1, frame, "vector")
        insert_waypoint_at_frame(animated_2, orig_path_2, frame, "vector")

    #################### END OF SECTION 1 ####################

    ### SECTION TRY ###
    # Every frames value is precomputed in order to achieve maximum similarity
    # to that of Synfig
    st_fr = min(get_frame(animated_1[0]), get_frame(animated_2[0]))
    en_fr = max(get_frame(animated_1[-1]), get_frame(animated_2[-1]))
    fra = st_fr
    while fra <= en_fr:
        insert_waypoint_at_frame(animated_1, orig_path_1, fra, "vector")
        insert_waypoint_at_frame(animated_2, orig_path_2, fra, "vector")
        fra += 1
    ### END SECTION ###

    ######################### SECTION 2 ##########################
    # Insert the waypoints at corresponding positions where point1 is animated
    # and point2 is animated
    for waypoint in animated_1:
        frame = get_frame(waypoint)
        insert_waypoint_at_frame(animated_2, orig_path_2, frame, "vector")

    for waypoint in animated_2:
        frame = get_frame(waypoint)
        insert_waypoint_at_frame(animated_1, orig_path_1, frame, "vector")
    ##################### END OF SECTION 2 #######################

    ##################### SECTION 3 ##############################
    # Add the impact of expand parameter amount towards point1 and point2
    # parameter
    assert len(animated_1) == len(animated_2)
    for waypoint1, waypoint2 in zip(animated_1, animated_2):
        frame = get_frame(waypoint1)
        assert frame == get_frame(waypoint2)
        expand_amount = get_vector_at_frame(expand_path, frame)
        expand_amount = to_Synfig_axis(expand_amount, "real")

        pos1, pos2 = get_vector(waypoint1), get_vector(waypoint2)
        # Comparing the x-coordinates
        if pos1[0] > pos2[0]:
            pos1[0] += expand_amount
            pos2[0] -= expand_amount
        else:
            pos1[0] -= expand_amount
            pos2[0] += expand_amount
        # Comparing the y-coordinates
        if pos1[1] > pos2[1]:
            pos1[1] += expand_amount
            pos2[1] -= expand_amount
        else:
            pos1[1] -= expand_amount
            pos2[1] += expand_amount
        set_vector(waypoint1, pos1)
        set_vector(waypoint2, pos2)
    ##################### END OF SECTION 3 #######################

    #################### SECTION 4 #############################
    # Place waypoints at which the x and y cross each other/cross the extremas
    cross_list = get_cross_list(animated_1, animated_2, orig_path_1, orig_path_2)
    for frame in cross_list:
        insert_waypoint_at_frame(animated_1, orig_path_1, frame, "vector")
        insert_waypoint_at_frame(animated_2, orig_path_2, frame, "vector")
    #################### END SECTION 4 #########################


    ################## SECTION 5 ################################################
    # Store the position of rectangle according to the waypoints in pos_animated
    # Store the size of rectangle according to the waypoints in size_animated
    pos_animated = copy.deepcopy(animated_1)
    size_animated = copy.deepcopy(animated_1)
    size_animated.attrib["type"] = "rectangle_size"

    i, i1 = 0, 0
    while i < len(animated_1) - 1:
        cur_get_after_1, cur_get_after_2 = animated_1[i].attrib["after"], animated_2[i].attrib["after"]
        next_get_before_1, next_get_before_2 = animated_1[i+1].attrib["before"], animated_2[i+1].attrib["before"]

        dic_1 = {"linear", "auto", "clamped", "halt"}
        dic_2 = {"constant"}
        constant_interval_1 = cur_get_after_1 in dic_2 or next_get_before_1 in dic_2
        constant_interval_2 = cur_get_after_2 in dic_2 or next_get_before_2 in dic_2

        # Case 1 no "constant" interval is present
        if (cur_get_after_1 in dic_1) and (cur_get_after_2 in dic_1) and (next_get_before_1 in dic_1) and (next_get_before_2 in dic_1):
            get_average(pos_animated[i1], animated_1[i], animated_2[i])
            copy_tcb_average(pos_animated[i1], animated_1[i], animated_2[i])

            get_difference(size_animated[i1], animated_1[i], animated_2[i])
            copy_tcb(size_animated[i1], pos_animated[i1])
            i, i1 = i + 1, i1 + 1
            get_average(pos_animated[i1], animated_1[i], animated_2[i])
            copy_tcb_average(pos_animated[i1], animated_1[i], animated_2[i])

            get_difference(size_animated[i1], animated_1[i], animated_2[i])
            copy_tcb(size_animated[i1], pos_animated[i1])

        # Case 2 only one "constant" interval: could mean two "constant"'s are present
        elif (constant_interval_1 and not constant_interval_2) or (not constant_interval_1 and constant_interval_2):
            if constant_interval_1:
                i, i1 = calc_pos_and_size(size_animated, pos_animated, animated_1, animated_2, orig_path_2, i, i1)
            elif constant_interval_2:
                i, i1 = calc_pos_and_size(size_animated, pos_animated, animated_2, animated_1, orig_path_1, i, i1)

        # Case 3 both are constant
        elif constant_interval_1 and constant_interval_2:
            # No need to copy tcb, as it's pos should be "constant"
            get_average(pos_animated[i1], animated_1[i], animated_2[i])
            get_difference(size_animated[i1], animated_1[i], animated_2[i])

            i, i1 = i + 1, i1 + 1
            get_difference(size_animated[i1], animated_1[i], animated_2[i])
            get_average(pos_animated[i1], animated_1[i], animated_2[i])
    ######################### SECTION 5 END ##############################

    ######################### SECTION 6 ##################################
    # Generate the position and size for lottie format
    gen_properties_multi_dimensional_keyframed(lottie["p"],
                                               pos_animated,
                                               index.inc())
    gen_value_Keyframed(lottie["s"], size_animated, index.inc())
    ########################## END OF SECTION 6 ###########################


def get_cross_list(animation_1, animation_2, orig_path_1, orig_path_2):
    """
    This function will return a list('set' technically) at which the point1 and point2 of rectangle
    will cross each other.
    This set might contain frames at which waypoints are already present, hence
    this need to be taken care of

    Args:
        animation_1 (lxml.etree._Element): Stores the animation of `point1` parameter in Synfig format
        animation_2 (lxml.etree._Element): Stores the animation of `point2` parameter in Synfig format
        orig_path_1 (dict)               : Stores the animation of `point1` parameter in Lottie format
        orig_path_2 (dict)               : Stores the animation of `point2` parameter in Lottie format

    Returns:
        (set) : Contains the frames at which point1 and point2 cross each other
    """
    en_fr = max(get_frame(animation_1[-1]), get_frame(animation_2[-1]))
    # Function to determine the sign of a variable
    sign = lambda a: (1, -1)[a < 0]
    prev_1 = float(animation_1[0][0][0].text), float(animation_1[0][0][1].text)
    prev_2 = float(animation_2[0][0][0].text), float(animation_2[0][0][1].text)

    # The list to be returned
    ret_list = set()
    frame = 1
    # Loop for all the frames
    while frame <= en_fr:
        now_1 = get_vector_at_frame(orig_path_1, frame)
        now_1 = to_Synfig_axis(now_1, "vector")
        now_2 = get_vector_at_frame(orig_path_2, frame)
        now_2 = to_Synfig_axis(now_2, "vector")
        is_needed = False
        if sign(prev_1[0] - prev_2[0]) != sign(now_1[0] - now_2[0]):
            is_needed = True
        elif sign(prev_1[1] - prev_2[1]) != sign(now_1[1] - now_2[1]):
            is_needed = True
        if is_needed:
            ret_list.add(frame - 1)
            ret_list.add(frame)
        prev_1, prev_2 = now_1, now_2
        frame += 1
    return ret_list


def calc_pos_and_size(size_animated, pos_animated, animated_1, animated_2, orig_path, i, i1):
    """
    Between two frames, this function is called if either "only point1's
    interval is constant" or "only point2's interval is constant". It calculates
    the position and size property for lottie format. It also adds a new
    waypoint just before the end of the frame if required.
    param3: Can be param1 or param2 of rectangle layer
    param4: Can be param2 or param1 of rectangle layer, but opposite of param3

    Args:
        size_animated (lxml.etree._Element): Holds the size parameter of rectangle layer in Synfig format
        pos_animated  (lxml.etree._Element): Holds the position parameter of rectangle layer in Synfig format
        animated_1    (lxml.etree._Element): Holds the param3 in Synfig format
        animated_2    (lxml.etree._Element): Holds the param4 in Synfig format
        orig_path     (dict)               : Holds the param4 in Lottie format
        i             (int)                : Iterator for animated_2
        i1            (int)                : Iterator for pos_animated and size_animated
    Returns:
        (int, int)    : Updated iterators i and i1 are returned
    """
    pos_animated[i1].attrib["after"] = animated_2[i].attrib["after"]
    size_animated[i1].attrib["after"] = animated_2[i].attrib["after"]

    copy_tcb(pos_animated[i1], animated_2[i])
    copy_tcb(size_animated[i1], animated_2[i])

    get_average(pos_animated[i1], animated_1[i], animated_2[i])
    get_difference(size_animated[i1], animated_1[i], animated_2[i])
    # Inserting a waypoint just before the nextwaypoint
    # Only if a waypoint can be inserted
    t_next = get_frame(animated_2[i+1])
    t_present = get_frame(animated_2[i])

    ######### Need to check if t_next - t_present < 2 #####
    if abs(t_next - t_present) >= 2:
        pos = get_vector_at_frame(orig_path, t_next - 1)
        pos = to_Synfig_axis(pos, "vector")
        new_waypoint = copy.deepcopy(pos_animated[i1])
        new_waypoint.attrib["before"] = new_waypoint.attrib["after"]
        new_waypoint.attrib["time"] = str((t_next - 1) / settings.lottie_format["fr"]) + "s"
        new_waypoint[0][0].text, new_waypoint[0][1].text = str(pos[0]), str(pos[1])

        n_size_waypoint = copy.deepcopy(new_waypoint)
        get_average(new_waypoint, new_waypoint, animated_1[i])
        get_difference(n_size_waypoint, n_size_waypoint, animated_1[i])

        pos_animated.insert(i1 + 1, new_waypoint)
        size_animated.insert(i1 + 1, n_size_waypoint)
        i1 += 1
    i, i1 = i + 1, i1 + 1
    get_average(pos_animated[i1], animated_1[i], animated_2[i])
    get_difference(size_animated[i1], animated_1[i], animated_2[i])
    return i, i1


def get_difference(waypoint, way_1, way_2):
    """
    Returns the absolute difference of vector's of way_1 and way_2 in "waypoint"
    Helpful in calculating 'size' parameter of lottie format from 'point1' and
    'point2' of Synfig format

    Args:
        waypoint (lxml.etree._Element) : Holds the waypoint at which the return value is to be stored
        way_1    (lxml.etree._Element) : Waypoint contributing in calculating absolute difference
        way_2    (lxml.etree._Element) : Waypoint contributing in calculating absolute difference

    Returns:
        (None)
    """
    waypoint[0].tag = "real" # If was vector before
    waypoint[0].attrib["value"] = str(abs(float(way_1[0][0].text) - float(way_2[0][0].text)))
    waypoint[0].attrib["value2"] = str(abs(float(way_1[0][1].text) - float(way_2[0][1].text)))


def get_average(waypoint, way_1, way_2):
    """
    Returns the average of vector's of way_1 and way_2 in "waypoint"
    Helpful in calculating 'position' parameter of lottie format from 'point1'
    and 'point2' of Synfig format

    Args:
        waypoint (lxml.etree._Element) : Holds the waypoint at which the return value is to be stored
        way_1    (lxml.etree._Element) : Waypoint contributing in calculating average
        way_2    (lxml.etree._Element) : Waypoint contributing in calculating average

    Returns:
        (None)
    """
    waypoint[0][0].text = str((float(way_1[0][0].text) + float(way_2[0][0].text)) / 2)
    waypoint[0][1].text = str((float(way_1[0][1].text) + float(way_2[0][1].text)) / 2)



"""
Will store all the function necessary for generation of offset properties of a
keyframe in lottie
"""

import sys
import copy
import settings
from common.misc import get_frame, parse_position, change_axis
from common.Vector import Vector
sys.path.append("..")


def isclose(a_val, b_val, rel_tol=1e-09, abs_tol=0.0):
    """
    A Function for testing approximate equality

    Args:
        a_val (float) : First value that needs equality testing
        b_val (float) : Second value that needs equality testing
        rel_tol (:obj: `float`, optional) : Relative tolerance
        abs_tol (:obj: `float`, optional) : Absolutte tolerance

    Returns:
        (bool) : Returns True If two numbers are equal according to specified tolerance
                         False Else
    """
    return abs(a_val - b_val) <= max(rel_tol * max(abs(a_val), abs(b_val)), abs_tol)


def clamped_tangent(p1, p2, p3, animated, i):
    """
    Function corresponding to clamped function in Synfig
    It generates the tangent when clamped waypoints are used

    Args:
        p1       (float)               : First point
        p2       (float)               : Second point
        p3       (float)               : Third point
        animated (lxml.etree._Element) : Synfig format animation
        i        (int)                 : Iterator over animation

    Returns:
        (float) : Clamped tangent is returned
    """
    # pw -> prev_waypoint, w -> waypoint, nw -> next_waypoint
    pw, w, nw = animated[i-1], animated[i], animated[i+1]
    t1 = get_frame(pw)
    t2 = get_frame(w)
    t3 = get_frame(nw)
    bias = 0.0
    tangent = 0.0
    pm = p1 + (p3 - p1)*(t2 - t1)/(t3 - t1)
    if p3 > p1:
        if p2 >= p3 or p2 <= p1:
            tangent = tangent * 0.0
        else:
            if p2 > pm:
                bias = (pm - p2) / (p3 - pm)
            elif p2 < pm:
                bias = (pm - p2) / (pm - p1)
            else:
                bias = 0.0
            tangent = (p2 - p1) * (1.0 + bias) / 2.0 + (p3 - p2) * (1.0 - bias) / 2.0
    elif p1 > p3:
        if p2 >= p1 or p2 <= p3:
            tangent = tangent * 0.0
        else:
            if p2 > pm:
                bias = (pm - p2) / (pm - p1)
            elif p2 < pm:
                bias = (pm - p2) / (p3 - pm)
            else:
                bias = 0.0
            tangent = (p2 - p1) * (1.0 + bias) / 2.0 + (p3 - p2) * (1.0 - bias) / 2.0
    else:
        tangent = tangent * 0.0
    return tangent


def clamped_vector(p1, p2, p3, animated, i, lottie, ease):
    """
    Function to generate the collective tangents i.e. x tangent and y tangent
    when clamped waypoints are used

    Args:
        p1       (common.Vector.Vector)         : First point in Coordinate System
        p2       (common.Vector.Vector)         : Second point in Coordinate System
        p3       (common.Vector.Vector)         : Third point in Coordinate System
        animated (lxml.etree._Element) : Synfig format animation
        i        (int)                 : Iterator over animation
        ease     (str)                 : Specifies if it is an ease in animation ease out

    Returns:
        (common.Vector.Vector) : Clamped Vector is returned
    """
    x_tan = clamped_tangent(p1[0], p2[0], p3[0], animated, i)
    y_tan = clamped_tangent(p1[1], p2[1], p3[1], animated, i)

    if isclose(x_tan, 0.0) or isclose(y_tan, 0.0):
        if ease == "in":
            ease_in(lottie)
        else:
            ease_out(lottie)
    return Vector(x_tan, y_tan, animated.attrib["type"])


def ease_out(lottie):
    """
    To set the ease out values in lottie format

    Args:
        lottie (dict) : Bezier interval of lottie

    Returns:
        (None)
    """
    lottie["o"]["x"] = settings.OUT_TANGENT_X
    lottie["o"]["y"] = settings.OUT_TANGENT_Y


def ease_in(lottie):
    """
    To set the ease in values in lottie format

    Args:
        lottie (dict) : Bezier interval of lottie

    Returns:
        (None)
    """
    lottie["i"]["x"] = settings.IN_TANGENT_X
    lottie["i"]["y"] = settings.IN_TANGENT_Y


def handle_color():
    """
    Default linear tangent values for color interpolations

    Args:
        (None)

    Returns:
        (common.Vector.Vector, common.Vector.Vector) : out and in tangents for color parameter
    """
    out_val = Vector(0.5, 0.5, "color")
    in_val = Vector(0.5, 0.5, "color")
    return out_val, in_val


def calc_tangent(animated, lottie, i):
    """
    Calculates the tangent, given two waypoints and there interpolation methods

    Args:
        animated (lxml.etree._Element) : Synfig format animation
        lottie   (dict)                : Lottie format animation stored here
        i        (int)                 : Iterator for animation

    Returns:
        (common.Vector.Vector) : If waypoint's value is parsed to common.Vector.Vector by misc.parse_position()
        (common.Color.Color)  : If waypoint's value is parsed to common.Color.Color ...
        (float)       : If waypoint's value is parsed to float ...
        (None)        : If "constant" interval is detected
    """
    waypoint, next_waypoint = animated[i], animated[i+1]
    cur_get_after, next_get_before = waypoint.attrib["after"], next_waypoint.attrib["before"]
    cur_get_before, next_get_after = waypoint.attrib["before"], next_waypoint.attrib["after"]

    if animated.attrib["type"] in {"angle", "star_angle_new", "region_angle"}:
    #if animated.attrib["type"] in {"angle"}:
        if cur_get_after == "auto":
            cur_get_after = "linear"
        if cur_get_before == "auto":
            cur_get_before = "linear"
        if next_get_before == "auto":
            next_get_before = "linear"
        if next_get_after == "auto":
            next_get_after = "linear"

    # Synfig only supports constant interpolations for points
    if animated.attrib["type"] == "points":
        cur_get_after = "constant"
        cur_get_before = "constant"
        next_get_after = "constant"
        next_get_before = "constant"

    # After effects only supports linear,ease-in,ease-out and constant interpolations for color
    ##### No support for TCB and clamped interpolations in color is there yet #####
    if animated.attrib["type"] in {"color", "gradient"}:
        if cur_get_after in {"auto", "clamped"}:
            cur_get_after = "linear"
        if cur_get_before in {"auto", "clamped"}:
            cur_get_before = "linear"
        if next_get_before in {"auto", "clamped"}:
            next_get_before = "linear"
        if next_get_after in {"auto", "clamped"}:
            next_get_after = "linear"

    # Calculate positions of waypoints
    cur_pos = parse_position(animated, i)
    prev_pos = copy.deepcopy(cur_pos)
    next_pos = parse_position(animated, i + 1)
    after_next_pos = copy.deepcopy(next_pos)

    if i + 2 <= len(animated) - 1:
        after_next_pos = parse_position(animated, i + 2)
    if i - 1 >= 0:
        prev_pos = parse_position(animated, i - 1)

    tens, bias, cont = 0, 0, 0   # default values
    tens1, bias1, cont1 = 0, 0, 0
    if "tension" in waypoint.keys():
        tens = float(waypoint.attrib["tension"])
    if "continuity" in waypoint.keys():
        cont = float(waypoint.attrib["continuity"])
    if "bias" in waypoint.keys():
        bias = float(waypoint.attrib["bias"])
    if "tension" in next_waypoint.keys():
        tens1 = float(next_waypoint.attrib["tension"])
    if "continuity" in next_waypoint.keys():
        cont1 = float(next_waypoint.attrib["continuity"])
    if "bias" in next_waypoint.keys():
        bias1 = float(next_waypoint.attrib["bias"])


    ### Special case for color interpolations ###
    if animated.attrib["type"] in {"color", "gradient"}:
        if cur_get_after == "linear" and next_get_before == "linear":
            return handle_color()

    # iter           next
    # ANY/TCB ------ ANY/ANY
    if cur_get_after == "auto":
        if i >= 1:
            out_val = ((1 - tens) * (1 + bias) * (1 + cont) *\
                       (cur_pos - prev_pos))/2 +\
                       ((1 - tens) * (1 - bias) * (1 - cont) *\
                       (next_pos - cur_pos))/2
        else:
            out_val = next_pos - cur_pos      # t1 = p2 - p1

    # iter           next
    # ANY/LINEAR --- ANY/ANY
    # ANY/EASE   --- ANY/ANY
    if cur_get_after in {"linear", "halt"}:
        out_val = next_pos - cur_pos

    # iter          next
    # ANY/ANY ----- LINEAR/ANY
    # ANY/ANY ----- EASE/ANY
    if next_get_before in {"linear", "halt"}:
        in_val = next_pos - cur_pos

    # iter          next
    # ANY/CLAMPED - ANY/ANY
    if cur_get_after == "clamped":
        if i >= 1:
            ease = "out"
            out_val = clamped_vector(prev_pos, cur_pos, next_pos, animated, i, lottie, ease)
        else:
            out_val = next_pos - cur_pos      # t1 = p2 - p1

    # iter          next             after_next
    # ANY/ANY ----- CLAMPED/ANY ---- ANY/ANY
    if next_get_before == "clamped":
        if i + 2 <= len(animated) - 1:
            ease = "in"
            in_val = clamped_vector(cur_pos,
                                    next_pos,
                                    after_next_pos,
                                    animated,
                                    i + 1,
                                    lottie,
                                    ease)
        else:
            in_val = next_pos - cur_pos     # t2 = p2 - p1

    # iter              next
    # ANY/CONSTANT ---- ANY/ANY
    # ANY/ANY      ---- CONSTANT/ANY
    if cur_get_after == "constant" or next_get_before == "constant":
        lottie["h"] = 1
        if animated.attrib["type"] == "vector":
            del lottie["to"], lottie["ti"]
        del lottie["i"], lottie["o"]
        # "e" is not needed, but is still not deleted as
        # it is of use in the last iteration of animation
        # See properties/multiDimenstionalKeyframed.py for more details
        # del lottie["e"]

        # If the number of points is decreasing, then hold interpolation should
        # have reverse effect. The value should instantly decrease and remain
        # same for the rest of the interval
        if animated.attrib["type"] == "points":
            if i > 0 and prev_pos[0] > cur_pos[0]:
                t_now = get_frame(animated[i-1]) + 1
                lottie["t"] = t_now
        return

    # iter           next           after_next
    # ANY/ANY ------ TCB/ANY ------ ANY/ANY
    if next_get_before == "auto":
        if i + 2 <= len(animated) - 1:
            in_val = ((1 - tens1) * (1 + bias1) * (1 - cont1) *\
                      (next_pos - cur_pos))/2 +\
                      ((1 - tens1) * (1 - bias1) * (1 + cont1) *\
                      (after_next_pos - next_pos))/2
        else:
            in_val = next_pos - cur_pos      # t2 = p2 - p1
    return out_val, in_val


def gen_properties_offset_keyframe(curve_list, animated, i):
    """
    Generates the dictionary corresponding to properties/offsetKeyFrame.json

    Args:
        curve_list (list)                : Stores bezier curve in Lottie format
        animated   (lxml.etree._Element) : Synfig format animation
        i          (int)                 : Iterator for animation

    Returns:
        (TypeError) : If a constant interval is encountered
        (None)      : In all other cases
    """
    lottie = curve_list[-1]

    waypoint, next_waypoint = animated[i], animated[i+1]
    cur_get_after, next_get_before = waypoint.attrib["after"], next_waypoint.attrib["before"]
    cur_get_before, next_get_after = waypoint.attrib["before"], next_waypoint.attrib["after"]

    # "angle" interpolations never call this function, can be removed by confirming
    if animated.attrib["type"] == "angle":
        if cur_get_after == "auto":
            cur_get_after = "linear"
        if cur_get_before == "auto":
            cur_get_before = "linear"
        if next_get_before == "auto":
            next_get_before = "linear"
        if next_get_after == "auto":
            next_get_after = "linear"

    # Synfig only supports constant interpolations for points
    # "points" never call this function, can be removed by confirming
    if animated.attrib["type"] == "points":
        cur_get_after = "constant"
        cur_get_before = "constant"
        next_get_after = "constant"
        next_get_before = "constant"

    # Calculate positions of waypoints
    cur_pos = parse_position(animated, i)
    next_pos = parse_position(animated, i + 1)

    lottie["i"] = {}    # Time bezier curve, not used in synfig
    lottie["o"] = {}    # Time bezier curve, not used in synfig
    lottie["i"]["x"] = 0.5
    lottie["i"]["y"] = 0.5
    lottie["o"]["x"] = 0.5
    lottie["o"]["y"] = 0.5
    if cur_get_after == "halt": # For ease out
        ease_out(lottie)
    if next_get_before == "halt": # For ease in
        ease_in(lottie)
    lottie["t"] = get_frame(waypoint)

    lottie["s"] = change_axis(cur_pos[0], cur_pos[1])
    lottie["e"] = change_axis(next_pos[0], next_pos[1])
    lottie["to"] = []
    lottie["ti"] = []

    # Calculating the unchanged tangent
    try:
        out_val, in_val = calc_tangent(animated, lottie, i)
    except Exception as excep:
        # This means constant interval
        return excep

    # This module is only needed for origin animation
    lottie["to"] = out_val.get_list()
    lottie["ti"] = in_val.get_list()

    # TCB/!TCB and list is not empty
    if cur_get_before == "auto" and cur_get_after != "auto" and i > 0:
        curve_list[-2]["ti"] = copy.deepcopy(lottie["to"])
        curve_list[-2]["ti"] = [-item/settings.TANGENT_FACTOR for item in curve_list[-2]["ti"]]
        curve_list[-2]["ti"][1] = -curve_list[-2]["ti"][1]
        if cur_get_after == "halt":
            curve_list[-2]["i"]["x"] = settings.IN_TANGENT_X
            curve_list[-2]["i"]["y"] = settings.IN_TANGENT_Y

    # Lottie tangent length is larger than synfig
    lottie["ti"] = [item/settings.TANGENT_FACTOR for item in lottie["ti"]]
    lottie["to"] = [item/settings.TANGENT_FACTOR for item in lottie["to"]]

    # Lottie and synfig use different tangents SEE DOCUMENTATION
    lottie["ti"] = [-item for item in lottie["ti"]]

    # IMPORTANT to and ti have to be relative
    # The y-axis is different in lottie
    lottie["ti"][1] = -lottie["ti"][1]
    lottie["to"][1] = -lottie["to"][1]

    # These tangents will be used in actual calculation of points according to
    # Synfig
    lottie["synfig_to"] = [tangent for tangent in lottie["to"]]
    lottie["synfig_ti"] = [-tangent for tangent in lottie["ti"]]
    if cur_get_after == "halt":
        lottie["synfig_to"] = [0 for val in lottie["synfig_to"]]
    if next_get_before == "halt":
        lottie["synfig_ti"] = [0 for val in lottie["synfig_ti"]]



"""
Will store all functions needed to generate the circle layer in lottie
This will also support the simple_circle layer of Synfig
"""

import sys
import settings
sys.path.append("..")


def gen_shapes_circle(lottie, layer, idx):
    """
    Generates the dictionary corresponding to shapes/ellipse.json where ellipse
    will always be considered as circle

    Args:
        lottie (dict)               : The lottie generated circle layer will be stored in it
        layer  (common.Layer.Layer)  : Synfig format circle layer
        idx    (int)                : Stores the index of the circle layer

    Returns:
        (None)
    """
    lottie["ty"] = "el"     # Type: circle
    lottie["p"] = {}        # Position of circle
    lottie["d"] = settings.DEFAULT_DIRECTION
    lottie["s"] = {}        # Size of circle
    lottie["ix"] = idx      # setting the index

    # Radius
    radius = layer.get_param("radius")
    radius.scale_convert_link(2)        # Lottie format expects diameter of the circle as input
    radius.animate("circle_radius")
    radius.fill_path(lottie, "s")

    # Origin
    origin = layer.get_param("origin", "center")
    origin.animate("vector")
    origin.fill_path(lottie, "p")



"""
This module will store all the functions required for fill effects of lottie
"""

import sys
import settings
from common.Count import Count
from effects.fillmask import gen_effects_fillmask
from effects.allmask import gen_effects_allmask
from effects.color import gen_effects_color
from effects.invert import gen_effects_invert
from effects.feather import gen_effects_hfeather, gen_effects_vfeather
from effects.opacity import gen_effects_opacity
sys.path.append("../")


def gen_effects_fill(lottie, layer, idx):
    """
    Generates the dictionary corresponding to effects/fill.json

    Args:
        lottie (dict)               : Lottie format layer
        layer  (common.Layer.Layer)  : Synfig format layer
        idx    (int)                : Index/Count of effect

    Returns:
        (None)
    """
    index = Count()
    lottie["ty"] = settings.EFFECTS_FILL    # Effect type
    lottie["nm"] = "Fill"                   # Name
    lottie["ix"] = idx                      # Index
    lottie["ef"] = []                       # Effect list of properties

    # generating the fill mask, has no use in Synfig. But a necessity for
    # running the .json file
    lottie["ef"].append({})
    gen_effects_fillmask(lottie["ef"][-1], index.inc())

    # generating the all mask property as required by lottie
    lottie["ef"].append({})
    gen_effects_allmask(lottie["ef"][-1], index.inc())

    # generating the color property
    lottie["ef"].append({})
    gen_effects_color(lottie["ef"][-1], layer, index.inc())

    # generating the invert property as required by lottie
    lottie["ef"].append({})
    gen_effects_invert(lottie["ef"][-1], index.inc())

    # generating the horizontal feather as required by lottie
    lottie["ef"].append({})
    gen_effects_hfeather(lottie["ef"][-1], index.inc())

    # generating the vertical feather as required by lottie
    lottie["ef"].append({})
    gen_effects_vfeather(lottie["ef"][-1], index.inc())

    # generating the opacity
    lottie["ef"].append({})
    gen_effects_opacity(lottie["ef"][-1], layer, index.inc())



"""
Will have all the functions required for generate the fill(color) in lottie
"""

import sys
from lxml import etree
import settings
from properties.value import gen_properties_value
from properties.valueKeyframed import gen_value_Keyframed
from common.misc import is_animated, real_high_precision
from common.Count import Count
from common.Gradient import Gradient
sys.path.append("..")


def gen_radial_gradient(lottie, layer, idx):
    """
    Generates the dictionary corresponding to shapes/gFill.json but with radial gradient

    Args:
    """
    index = Count()
    lottie["ty"] = "gf"
    lottie["r"] = 1    # Don't know it's meaning yet, but works without this also
    lottie["o"] = {}   # Opacity of the gradient layer
    lottie["nm"] = layer.get_description()
    lottie["t"] = 2    # 2 means radial gradient layer
    lottie["s"] = {}   # Starting point of gradient
    lottie["e"] = {}   # Ending point of gradient
    lottie["g"] = {}   # Gradient information is stored here

    # Color Opacity
    opacity = layer.get_param("amount").get()
    is_animate = is_animated(opacity[0])
    if is_animate == settings.ANIMATED:
        # Telling the function that this is for opacity
        opacity[0].attrib['type'] = 'opacity'
        gen_value_Keyframed(lottie["o"], opacity[0], index.inc())

    else:
        if is_animate == settings.NOT_ANIMATED:
            val = float(opacity[0].attrib["value"]) * settings.OPACITY_CONSTANT
        else:
            val = float(opacity[0][0][0].attrib["value"]) * settings.OPACITY_CONSTANT
        gen_properties_value(lottie["o"],
                             val,
                             index.inc(),
                             settings.DEFAULT_ANIMATED,
                             settings.NO_INFO)

        # Gradient colors
        lottie["g"]["k"] = {}
        lottie["g"]["ix"] = index.inc()
        gradient = layer.get_param("gradient")
        modify_gradient(gradient)
        gradient.animate("gradient")  # To find the lottie path of the modified gradient
        lottie["g"]["p"] = len(gradient.get()[0][0][0])
        gradient.fill_path(lottie["g"], "k")
        modify_gradient_according_to_latest_version(lottie["g"]["k"])

        # Starting point and Ending points need to be retrieved from center and radius
        center = layer.get_param("center")
        center.animate("vector")
        center.fill_path(lottie, "s")

        radius = layer.get_param("radius")
        radius.animate("real")

        # Ending point will be (start[0] + radius, start[1])
        # Below is just a modification of fill_path function
        expression = "var $bm_rt; $bm_rt = {expr}"
        x_expr = "sum(" + center.expression + "[0], " + radius.expression + ")"
        y_expr = center.expression + "[1]"
        expr = "[" + x_expr + ", " + y_expr + "]"
        expression = expression.format(expr=expr)
        gen_properties_value(lottie["e"],
                             [1, 1],
                             0,
                             0,
                             expression)
        if "ef" not in center.get_layer().get_lottie_layer().keys():
                center.get_layer().get_lottie_layer()["ef"] = []
        # If center has any expression controllers, then they would have been pushed earlier by fill_path, hence no need
        # center.get_layer().get_lottie_layer()["ef"].extend(center.expression_controllers)
        center.get_layer().get_lottie_layer()["ef"].extend(radius.expression_controllers)



def gen_linear_gradient(lottie, layer, idx):
    """
    Generates the dictionary corresponding to shapes/gFill.json but with linear gradient

    Args:
    """
    index = Count()
    lottie["ty"] = "gf"
    lottie["r"] = 1    # Don't know it's meaning yet, but works without this also
    lottie["o"] = {}   # Opacity of the gradient layer
    lottie["nm"] = layer.get_description()
    lottie["t"] = 1    # 1 means linear gradient layer
    lottie["s"] = {}   # Starting point of gradient
    lottie["e"] = {}   # Ending point of gradient
    lottie["g"] = {}   # Gradient information is stored here

    # Color Opacity
    opacity = layer.get_param("amount").get()
    is_animate = is_animated(opacity[0])
    if is_animate == settings.ANIMATED:
        # Telling the function that this is for opacity
        opacity[0].attrib['type'] = 'opacity'
        gen_value_Keyframed(lottie["o"], opacity[0], index.inc())

    else:
        if is_animate == settings.NOT_ANIMATED:
            val = float(opacity[0].attrib["value"]) * settings.OPACITY_CONSTANT
        else:
            val = float(opacity[0][0][0].attrib["value"]) * settings.OPACITY_CONSTANT
        gen_properties_value(lottie["o"],
                             val,
                             index.inc(),
                             settings.DEFAULT_ANIMATED,
                             settings.NO_INFO)

        # Starting point
        p1 = layer.get_param("p1")
        p1.animate("vector")
        p1.fill_path(lottie, "s")

        # Ending point
        p2 = layer.get_param("p2")
        p2.animate("vector")
        p2.fill_path(lottie, "e")

        # Gradient colors
        lottie["g"]["k"] = {}
        lottie["g"]["ix"] = index.inc()
        gradient = layer.get_param("gradient")
        modify_gradient(gradient)
        gradient.animate("gradient")  # To find the lottie path of the modified gradient
        lottie["g"]["p"] = len(gradient.get()[0][0][0])
        gradient.fill_path(lottie["g"], "k")
        modify_gradient_according_to_latest_version(lottie["g"]["k"])


def modify_gradient_according_to_latest_version(lottie):
    """
    Modifies the gradient's "g" property to put an "s" value at each keyframe instead of having "s" & "e"
    both.
    Reference: https://github.com/airbnb/lottie-web/issues/2055#issuecomment-605455945

    Args:
        lottie (dict) : Lottie format gradient colors

    Returns:
        (None)
    """
    for i in range(len(lottie["k"]) - 1, -1, -1):
        if i != 0:
            lottie["k"][i]["s"] = lottie["k"][i-1]["e"]
        # temp = lottie["k"][i].pop("e", "No key found")


def modify_gradient(gradient):
    """
    Finds the number of different positions where the color is defined in a gradient layer, and adds the color
    to each of those position for every waypoint

    Args:
        gradient (common.Param.Param) : Synfig format gradient layer

    Returns:
        (None)
    """
    positions = []
    gradient.animate("gradient")    # This is called initially so as to ensure that each color is inside a waypoint 
    for waypoint in gradient.get()[0]:
        gd = waypoint[0]
        for color in gd:
            positions.append(float(color.attrib["pos"]))
    positions.sort()

    # Synfig uses real_high_precision to differentiate 2 items
    discard = set()
    for i in range(len(positions)):
        for j in range(i+1, len(positions)):
            if abs(positions[i] - positions[j]) < real_high_precision():
                discard.add(j)
    discard_values = [positions[i] for i in discard]

    for val in discard_values:
        positions.remove(val)

    # Now the color is added to all the positions in all the waypoints
    for waypoint in gradient.get()[0]:
        gd = Gradient(waypoint[0])
        add_colors_to_gradient(waypoint, gd, positions)

    
def add_colors_to_gradient(waypoint, gd, positions):
    """
    Deletes previous values of color and introduce new depending upon the positions given

    Args:
        waypoint () : 
        gd () :

    Returns:
        (None)
    """
    # First remove all the colors, easy to remove all first and then add all+some additionals back
    for col in waypoint[0]:
        col.getparent().remove(col)

    # color template
    st = "<color pos='{pos}'><r>{red}</r><g>{green}</g><b>{blue}</b><a>{alpha}</a></color>"

    # Now add all the colors back
    for val in positions:
        color = gd.get_color_at_x(val)
        lxml_col = etree.fromstring(st.format(pos=val, red=color.red, green=color.green, blue=color.blue, alpha=color.alpha))
        waypoint[0].append(lxml_col)



# pylint: disable=line-too-long
"""
Matrix2.py
Will store class for a Matrix 2X2
"""

import sys
import math
from common.Vector import Vector
sys.path.append("..")


class Matrix2:
    """
    Will store a Matrix 2X2
    """
    def __init__(self, m00=1.0, m01=0.0, m10=0.0, m11=1.0):
        """
        Args:
            m00 (:obj: `float`, optional) : 0th row, 0th column value
            m01 (:obj: `float`, optional) : 0th row, 1st column value
            m10 (:obj: `float`, optional) : 1st row, 0th column value
            m11 (:obj: `float`, optional) : 1st row, 1st column value

        Returns:
            (None)
        """
        self.m00 = m00
        self.m01 = m01
        self.m10 = m10
        self.m11 = m11

    def set_rotate(self, angle):
        """
        This will serve as a rotation matrix with angle `angle`

        Args:
            angle (float) : Angle in radians

        Returns:
            (None)
        """
        c = math.cos(angle)
        s = math.sin(angle)
        self.m00, self.m01 = c, s
        self.m10, self.m11 = -s, c

    def get_transformed(self, v):
        """
        Rotate or transform a vector using this Matrix

        Args:
            v (common.Vector.Vector) : Vector to be rotated using this matrix

        Returns:
            (common.Vector.Vector) : Rotated vector
        """
        ret = Vector()
        x, y = v[0], v[1]
        ret[0] = x*self.m00 + y*self.m10
        ret[1] = x*self.m01 + y*self.m11
        return ret



"""
Will store all the functions corresponding to shapes in lottie
"""

import sys
import settings
from common.Count import Count
from shapes.star import gen_shapes_star
from shapes.circle import gen_shapes_circle
from shapes.fill import gen_shapes_fill
from shapes.rectangle import gen_shapes_rectangle
from shapes.shape import gen_shapes_shape
from shapes.gFill import gen_linear_gradient, gen_radial_gradient
from helpers.blendMode import get_blend
from helpers.transform import gen_helpers_transform
from synfig.rectangle import gen_dummy_rectangle
from properties.shapePropKeyframe.constant_width_outline import gen_bline_outline_constant
sys.path.append("..")


def gen_layer_shape(lottie, layer, idx):
    """
    Generates the dictionary corresponding to layers/shape.json

    Args:
        lottie (dict)       : Lottie generate shape stored here
        layer  (common.Layer.Layer) : Synfig format shape layer
        idx    (int)        : Stores the index(number of) of shape layer

    Returns:
        (None)
    """

    if layer.get_type() in {"linear_gradient", "radial_gradient"}:
        # Create dummy point1 and point2 for linear/radial gradient to generate rectangle for it in lottie format
        # Will have to use add_offset() function inside after this generation
        gen_dummy_rectangle(layer)

    layer.add_offset()

    index = Count()
    lottie["ddd"] = settings.DEFAULT_3D
    lottie["ind"] = idx
    lottie["ty"] = settings.LAYER_SHAPE_TYPE
    lottie["nm"] = layer.get_description()
    lottie["sr"] = settings.LAYER_DEFAULT_STRETCH
    lottie["ks"] = {}   # Transform properties to be filled

    gen_helpers_transform(lottie["ks"])

    lottie["ao"] = settings.LAYER_DEFAULT_AUTO_ORIENT
    lottie["shapes"] = []   # Shapes to be filled yet
    lottie["shapes"].append({})
    if layer.get_type() == "star":
        gen_shapes_star(lottie["shapes"][0], layer, index.inc())
    elif layer.get_type() in {"circle", "simple_circle"}:
        gen_shapes_circle(lottie["shapes"][0], layer, index.inc())
    elif layer.get_type() in {"filled_rectangle", "rectangle"}:
        gen_shapes_rectangle(lottie["shapes"][0], layer.get_layer(), index.inc())
    elif layer.get_type() in {"linear_gradient", "radial_gradient"}:
        gen_shapes_shape(lottie["shapes"][0], layer, index.inc())
    elif layer.get_type() in {"outline"}:
        settings.OUTLINE_FLAG = True
        gen_bline_outline_constant(lottie["shapes"][0],layer.get_param("bline"),layer,lottie["ks"],index.inc())

    lottie["shapes"].append({})  # For the fill or color
    if layer.get_type() in {"linear_gradient"}:
        gen_linear_gradient(lottie["shapes"][1], layer, index.inc())
    elif layer.get_type() in {"radial_gradient"}:
        gen_radial_gradient(lottie["shapes"][1], layer, index.inc())
    elif layer.get_type() in {"outline"}:
        pass
    else:
        gen_shapes_fill(lottie["shapes"][1], layer)

    lottie["ip"] = settings.lottie_format["ip"]
    lottie["op"] = settings.lottie_format["op"]
    lottie["st"] = 0            # Don't know yet
    get_blend(lottie, layer)



# pylint: disable=line-too-long
"""
Hermite.py
This will store the class Hermite which will store a Hermite curve
"""

import sys
import math
sys.path.append("..")


class Hermite:
    """
    To keep the hermite curve stored inside this
    """
    def __init__(self, p1, p2, t1, t2, r = 0.0, s = 1.0):
        """
        Args:
            p1 (float)  : First control point of bezier curve
            p2 (float)  : Fourth control point
            t1 (float)  : Second control point
            t2 (float)  : Third control point

        Returns:
            (None)
        """
        self.p1 = p1
        self.p2 = p2
        self.t1 = t1
        self.t2 = t2
        self.r = r
        self.s = s
        self.drs = 1.0 / (s - r)
        self.sync()

    def sync(self):
        """
        Calculates the coefficients and variables required in calculation of
        derivative and value of the curve at time t.
        Here p1, p2, t1, t2 are control points of bezier curve and a, b, c, d
        are control points of hermite

        Args:
            (None)
        Returns:
            (None)
        """
        self.a = self.p1
        self.b = self.p1 + self.t1/3
        self.c = self.p2 - self.t2/3
        self.d = self.p2

        self.coeff0 = self.a
        self.coeff1 = self.b*3- self.a*3
        self.coeff2 = self.c*3 - self.b*6 + self.a*3
        self.coeff3 = self.d - self.c*3 + self.b*3 - self.a

    def derivative(self, x):
        """
        Calculates the derivative of the curve at x

        Args:
            x (float) : Time at which the derivative is needed

        Returns:
            (float) : The value of the derivative at time x
        """
        y = 1 - x
        ret = ((self.b - self.a) * y * y + (self.c - self.b) * x * y * 2 + (self.d - self.c) * x * x) * 3
        return ret

    def value(self, t):
        """
        Calculates the value of the curve at time t

        Args:
            t (float) : Time at which the value is needed

        Returns:
            (float) : Value of the curve at time t
        """
        ret = self.coeff0 + (self.coeff1 + (self.coeff2 + (self.coeff3)*t)*t)*t
        return ret

    def find_distance(self, r, s, steps = 7):
        """
        https://github.com/synfig/synfig/blob/15607089680af560ad031465d31878425af927eb/ETL/ETL/_bezier.h#L577

        Args:
            r (float) : Time type
            s (float) : Time type

        Returns:
            (float) : The distance according to the Synfig link
        """
        inc = (s - r) / steps
        if not inc:
            return 0
        ret = 0.0

        last = self.single_operator(r)

        r += inc
        while r < s:
            n = self.single_operator(r)
            ret += self.uncook(self.double_operator(last, n))
            last = n
            r += inc

        ret += self.uncook(self.double_operator(last, self.single_operator(r)))*(s-(r-inc))/inc
        return ret

    def single_operator(self, t):
        """
        https://github.com/synfig/synfig/blob/15607089680af560ad031465d31878425af927eb/ETL/ETL/_bezier.h#L337

        This might differ, check if the results are wrong, then try this out: https://github.com/synfig/synfig/blob/15607089680af560ad031465d31878425af927eb/ETL/ETL/_bezier.h#L96

        Args:
            t (Float)

        Returns:
            (Float)
        """
        t -= self.r
        t *= self.drs

        return self.coeff0 + (self.coeff1 + (self.coeff2 + (self.coeff3)*t)*t)*t

    def double_operator(self, a, b):
        """
        https://github.com/synfig/synfig/blob/15607089680af560ad031465d31878425af927eb/ETL/ETL/_curve_func.h#L56

        Args:
            a (Template T):
            b (Template T):

        Returns:
            (T)
        """
        delta = b - a
        return delta*delta

    def uncook(self, x):
        """
        https://github.com/synfig/synfig/blob/15607089680af560ad031465d31878425af927eb/ETL/ETL/_curve_func.h#L63

        Args:
            x (float)

        Returns:
            (float)
        """
        return math.sqrt(x)

    def cook(self, x):
        """
        https://github.com/synfig/synfig/blob/15607089680af560ad031465d31878425af927eb/ETL/ETL/_curve_func.h#L62

        Args:
            x (Template T)

        Returns:
            (T)
        """
        return x*x

    def length(self):
        """
        https://github.com/synfig/synfig/blob/15607089680af560ad031465d31878425af927eb/ETL/ETL/_bezier.h#L595

        Args:
            (None)

        Returns:
            (float)
        """
        return self.find_distance(self.r, self.s)



# pylint: disable=line-too-long
"""
Will store all the helper functions and modules for generation of shapPropKeyframe file
in Lottie format
"""

import sys
import settings
from common.misc import change_axis, radial_to_tangent
from common.Vector import Vector
from common.Param import Param
sys.path.append("../../")


def animate_tangents(tangent, window):
    """
    Animates the radial composite and updates the window of frame if radial
    composite's parameters are already animated
    Also generate the Lottie path and stores in radial_composite

    Args:
        tangent (common.Param.Param)  : Synfig format radial composite's parent-> stores radius and angle
        window  (dict)                : max and min frame of overall animations stored in this

    Returns:
        (None)
    """
    for child in tangent[0]:
        # Assuming tangent[0] is always radial_composite
        if child.tag == "radius":
            radius = Param(child, tangent[0])
        elif child.tag == "theta":
            theta = Param(child, tangent[0])
    radius.update_frame_window(window)
    theta.update_frame_window(window)

    radius.animate("real")
    theta.animate("region_angle")

    tangent.add_subparam("radius", radius)
    tangent.add_subparam("theta", theta)


def update_child_at_parent(parent, new_child, tag, param_name=None):
    """
    Given a node, replaces the child with tag `tag` with new_child

    Args:
        parent    (lxml.etree._Element) : Node whose child needs to be replaced
        new_child (lxml.etree._Element) : Updated child of the parent
        tag       (str)                 : To identify the appropriate child

    Returns:
        (None)
    """
    if not param_name:
        for chld in parent:
            if chld.tag == tag:
                chld.getparent().remove(chld)
    else:
        for chld in parent:
            if chld.tag == tag and chld.attrib["name"] == param_name:
                chld.getparent().remove(chld)
    parent.insert(0, new_child)


def get_tangent_at_frame(t1, t2, fr):
    """
    Given a frame, returns the in-tangent and out-tangent at a bline point
    depending on whether split_radius and split_angle is "true"/"false"

    Args:
        t1      (common.Param.Param)  : Holds Tangent 1/In Tangent
        t2      (common.Param.Param)  : Holds Tangent 2/Out Tangent
        split_r (common.Param.Param) : Holds animation of split radius parameter
        split_a (common.Param.Param) : Holds animation of split angle parameter
        fr      (int)                 : Holds the frame value

    Returns:
        (common.Vector.Vector, common.Vector.Vector) : In-tangent and out-tangent at the given frame
    """

    # Setting tangent 1
    r1 = t1.get_subparam("radius").get_value(fr)
    a1 = t1.get_subparam("theta").get_value(fr)

    x, y = radial_to_tangent(r1, a1)
    tangent1 = Vector(x, y)

    # Setting tangent 2
    r2 = t2.get_subparam("radius").get_value(fr)
    a2 = t2.get_subparam("theta").get_value(fr)

    x, y = radial_to_tangent(r2, a2)
    tangent2 = Vector(x, y)

    return tangent1, tangent2


def add(side, lottie, origin_cur, is_rectangle=False):
    """
    Does not take care of tangent putting order because the tangents are all
    zero for now according to the code

    Args:
        side (list) : Stores the newly calculated vertex and tangents of outline layer
        lottie (dict) : These vertices will be stored in this dictionary
        origin_cur (list) : Value of origin at specific frame

    Returns:
        (None)
    """
    i = 0
    while i < len(side):
        cubic_to(side[i][0], side[i][1], side[i][2], lottie, origin_cur, is_rectangle)
        i += 1


def add_reverse(side, lottie, origin_cur):
    """
    Does not take care of tangents putting order because the tangents are all
    zero for now according to the code:

    Args:
        side (list) : Stores the newly calculated vertex and tangents of outline layer
        lottie (dict) : These vertices will be stored in this dictionary
        origin_cur (list) : Value of origin at specific frame

    Returns:
        (None)
    """
    i = len(side) - 1
    while i >= 0:
        cubic_to(side[i][0], side[i][1], side[i][2], lottie, origin_cur)
        i -= 1


def cubic_to(vec, tan1, tan2, lottie, origin_cur, is_rectangle=False,constant_width=False):
    """
    Will have to manipulate the tangents here, but they are not managed as tan1
    and tan2 both are zero always

    Args:
        vec (common.Vector.Vector) : position of the point
        tan1 (common.Vector.Vector) : tangent 1 of the point
        tan2 (common.Vector.Vector) : tangent 2 of the point
        lottie (dict) : Final position and tangents will be stored here
        origin_cur (list) : value of the origin at specific frame

    Returns:
        (None)
    """
    vec *= settings.PIX_PER_UNIT
    tan1 *= settings.PIX_PER_UNIT
    tan2 *= settings.PIX_PER_UNIT
    if constant_width:
        tan1, tan2 = convert_tangent_to_lottie(tan1, tan2)
    else:
        tan1, tan2 = convert_tangent_to_lottie(3*tan1, 3*tan2)
    pos = change_axis(vec[0], vec[1], not is_rectangle)
    for i in range(len(pos)):
        pos[i] += origin_cur[i]
    lottie["i"].append(tan1.get_list())
    lottie["o"].append(tan2.get_list())
    lottie["v"].append(pos)


def move_to(vec, lottie, origin_cur):
    """
    Don't have to manipulate the tangents because all of them are zero here

    Args:
        vec (common.Vector.Vector) : position of the point
        lottie (dict) : Final position and tangents will be stored here
        origin_cur (list) : value of the origin at specific frame

    Returns:
        (None)
    """
    vec *= settings.PIX_PER_UNIT
    lottie["i"].append([0, 0])
    lottie["o"].append([0, 0])
    pos = change_axis(vec[0], vec[1])
    for i in range(len(pos)):
        pos[i] += origin_cur[i]
    lottie["v"].append(pos)


def convert_tangent_to_lottie(t1, t2):
    """
    Converts tangent from Synfig format to lottie format

    Args:
        t1 (common.Vector.Vector) : tangent 1 of a point
        t2 (common.Vector.Vector) : tangent 2 of a point

    Returns:
        (common.Vector.Vector) : Converted tangent 1
        (common.Vector.Vector) : Converted tangent 2
    """
    # Convert to Lottie format
    t1 /= 3
    t2 /= 3

    # Synfig and Lottie use different in tangents SEE DOCUMENTATION
    t1 *= -1

    # Important: t1 and t2 have to be relative
    # The y-axis is different in lottie
    t1[1] = -t1[1]
    t2[1] = -t2[1]
    return t1, t2


def insert_dict_at(lottie, idx, fr, loop, constant=False):
    """
    Inserts dictionary values in the main dictionary, required by shape layer of
    lottie format

    Args:
        lottie (dict) : Shape layer will be stored in this main dictionary
        idx    (int)  : index at which dictionary needs to be stored
        fr     (int)  : frame number
        loop   (bool) : Specifies if the shape is loop or not
        constant(bool): Specifies if the user wants a constant shaped outline

    Returns:
        (dict) : Starting dictionary for shape interpolation
        (dict) : Ending dictionary for shape interpolation
    """
    if idx != -1:
        lottie.insert(idx, {})
    else:
        lottie.append({})
    lottie[idx]["i"], lottie[idx]["o"] = {}, {}
    lottie[idx]["i"]["x"] = lottie[idx]["i"]["y"] = 0.5     # Does not matter because frames are adjacent
    lottie[idx]["o"]["x"] = lottie[idx]["o"]["y"] = 0.5     # Does not matter because frames are adjacent
    lottie[-1]["t"] = fr
    if constant:
        lottie[idx]["s"] = []
        st_val = lottie[idx]["s"]

        st_val.append({})
        st_val = st_val[0]
        st_val["i"], st_val["o"], st_val["v"], st_val["c"] = [], [], [], loop
        return st_val
    else:
        lottie[idx]["s"], lottie[idx]["e"] = [], []
        st_val, en_val = lottie[idx]["s"], lottie[idx]["e"]

        st_val.append({}), en_val.append({})
        st_val, en_val = st_val[0], en_val[0]
        st_val["i"], st_val["o"], st_val["v"], st_val["c"] = [], [], [], loop
        en_val["i"], en_val["o"], en_val["v"], en_val["c"] = [], [], [], loop
        return st_val, en_val


def insert_dict_at_adv_outline(lottie, idx, fr, loop):
    """
    Inserts dictionary values in the main dictionary, required by shape layer of
    Lottie format. But this is specific for advanced outline layer, as we needed
    hold/constant interpolation between alternate frames

    Args:
        lottie (dict) : Shape layer will be stored in this main dictionary
        idx    (int)  : index at which dictionary needs to be stored
        fr     (int)  : frame number
        loop   (bool) : Specifies if the shape is loop or not

    Returns:
        (dict) : Starting dictionary for shape interpolation
        (dict) : Ending dictionary for shape interpolation

    """
    if idx != -1:
        lottie.insert(idx, {})
    else:
        lottie.append({})
    lottie[idx]["i"], lottie[idx]["o"] = {}, {}
    lottie[idx]["i"]["x"] = lottie[idx]["i"]["y"] = 0.5     # Does not matter because frames are adjacent
    lottie[idx]["o"]["x"] = lottie[idx]["o"]["y"] = 0.5     # Does not matter because frames are adjacent
    lottie[idx]["h"] = 1
    lottie[idx]["t"] = fr
    lottie[idx]["s"], lottie[idx]["e"] = [], []
    st_val, en_val = lottie[idx]["s"], lottie[idx]["e"]

    st_val.append({}), en_val.append({})
    st_val, en_val = st_val[0], en_val[0]
    st_val["i"], st_val["o"], st_val["v"], st_val["c"] = [], [], [], loop
    en_val["i"], en_val["o"], en_val["v"], en_val["c"] = [], [], [], loop
    return st_val, en_val

def quadratic_to_cubic(qp0, qp1, qp2):
    """
    Converts quadratic bezier curve to cubic bezier curve

    Args:
        qp0 (common.Vector.Vector) First control point of quadratic bezier
        qp1 (common.Vector.Vector) Second control point of quadratic bezier
        qp2 (common.Vector.Vector) Third control point of quadratic bezier

    Returns:
        (common.Vector.Vector) Second control point of Cubic bezier
        (common.Vector.Vector) Third control point of Cubic bezier
    """
    cp0 = qp0
    cp3 = qp2
    cp1 = qp0 + 2/3.0*(qp1 - qp0)
    cp2 = qp2 + 2/3.0*(qp1 - qp2)
    return cp1, cp2



"""
Param.py
Will store the Parameters class for Synfig parameters 
"""

import sys
import copy
import math
from lxml import etree
import settings
import common
import synfig.group
from synfig.animation import modify_bool_animation, to_Synfig_axis, is_animated, get_bool_at_frame, get_vector_at_frame
from properties.multiDimensionalKeyframed import gen_properties_multi_dimensional_keyframed
from properties.valueKeyframed import gen_value_Keyframed
from properties.value import gen_properties_value
from effects.controller import gen_effects_controller
sys.path.append("..")


class Param:
    """
    Class to keep Synfig format parameters
    """
    def __init__(self, param, parent):
        """
        This parent can be another parameter or a Layer or Canvas
        """
        self.parent = parent
        self.param = param
        self.subparams = {}
        self.SUBPARAMS_EXTRACTED = 0
        self.expression_controllers = [] # Effects will be stored in this
        self.expression = ""
        self.dimension = 1  # 1 represents real, 2 represents vector
        self.get_exported_valuenode()

        self.is_group_child = 0

    def get_exported_valuenode(self):
        """
        If the animation of this param is not directly linked with this param,
        then it must be preset in the canvas :use, :def
        """
        if self.param is None or "use" not in self.param.attrib.keys():
            return

        # Full ID string format is like: [[file-path]#][canvas-id:][child-canvas-id:][child-canvas-id:]…value-node-id
        # Refer here: https://github.com/synfig/synfig/issues/993
        layer = self.get_layer()
        canvas = layer.getparent()
        key = self.param.attrib["use"]
        keys = key.split("#")

        if len(keys) == 2:  # Meaning file-path is not empty
            root = etree.parse(keys[0]).getroot()
            # This is a hack which changes the actual root canvas, so that we
            # can store the root canvas of another file in our settings
            actual_root_canvas = settings.ROOT_CANVAS
            canvas = common.Canvas.Canvas(root, True)
            settings.ROOT_CANVAS = actual_root_canvas

            keys = keys[1].split(":")
        else:
            if ":" in keys[0]:  # Start searching from root canvas
                canvas = settings.ROOT_CANVAS
            keys = keys[0].split(":")   # Split based on ":" to go to the child-canvas-id

        # Iterate on cavas id's to reach the value node id
        for i in range(0, len(keys) - 1):
            if keys[i] == '':
                continue
            canvas = common.Canvas.Canvas(canvas.get_def(keys[i]))
        anim = canvas.get_def(keys[-1])
        assert(anim is not None)
        self.param.append(copy.deepcopy(anim))
        
    def reset(self):
        """
        This function will be called when any of the entities of the parameter
        are changed. This should be internally called, but for now compromising
        """
        self.PATH_GENERATED = 0
        self.TRANSFORM_PATH_GENERATED = 0
        self.path = {}

    def get(self):
        """
        Returns the original param
        """
        return self.param

    def get_text(self):
        """
        Returns the text stored inside
        """
        return self.param.text

    def __getitem__(self, itr):
        """
        Returns the child corresponding to itr
        """
        return self.param[itr]

    def __setitem__(self, itr, val):
        """
        Sets the value of child corresponding to itr
        """
        self.param[itr] = val

    def get_layer_type(self):
        """
        Recursively go till the top until layer is not reached
        """
        if isinstance(self.parent, common.Layer.Layer):
            return self.parent.get_type()
        return self.parent.get_layer_type()

    def get_layer(self):
        """
        Recursively find the layer
        """
        if isinstance(self.parent, common.Layer.Layer):
            return self.parent
        return self.parent.get_layer()

    def get_canvas(self):
        """
        Recursively find the canvas
        """
        if isinstance(self.parent, common.Canvas.Canvas):
            return self.parent
        return self.parent.get_canvas()

    def getparent(self):
        """
        Returns the parent of this parameter
        """
        return self.parent

    def add_subparam(self, key, val):
        """
        Adds sub parameters
        """
        self.subparams[key] = val

    def get_subparam_dict(self):
        """
        Returns the address of the dict
        """
        return self.subparams

    def get_subparam(self, key):
        """
        Given a key, returns it's dictionary
        """
        return self.subparams[key]

    def append_same_key(self, key, child):
        """
        Helps append in the dict if same key is already present
        """
        val = Param(child, self)

        if self.is_group_child:
            val.is_group_child = True

        if isinstance(self.subparams[key], list):
            self.subparams[key].append(val)
        else:
            prev = self.subparams[key]
            self.subparams[key] = [prev, val]

    def extract_subparams(self):
        """
        Extracts the subparameters of this parameter and stores with there tag
        as the key
        """
        if self.SUBPARAMS_EXTRACTED:
            return
        self.SUBPARAMS_EXTRACTED = 1
        self.subparams.clear()
        for child in self.param:
            key = child.tag 
            if key in self.subparams.keys():
                self.append_same_key(key, child)
            else:
                self.subparams[key] = Param(child, self)

                if self.is_group_child:
                    self.subparams[key].is_group_child = True

    def animate_without_path(self, anim_type):
        """
        If this parameter is not animated, it generates dummy waypoints and
        animates this parameter
        """
        if anim_type in {"vector", "group_layer_scale", "stretch_layer_scale", "circle_radius"}:   # This will never happen, can remove this latter
            self.dimension = 2

        # Check if we are dealing with convert methods
        if self.param[0].tag in settings.CONVERT_METHODS:
            self.extract_subparams()
            if self.param[0].tag == "add":
                self.subparams["add"].extract_subparams()
                self.subparams["add"].subparams["lhs"].animate(anim_type)
                self.subparams["add"].subparams["rhs"].animate(anim_type)
                self.subparams["add"].subparams["scaler"].animate("scalar_multiply")
        else:
            self.single_animate(anim_type)


    def animate(self, anim_type, transform=False):
        """
        If this parameter is not animated, it generates dummy waypoints and
        animates this parameter
        """
        if not transform and anim_type in {"vector"}:
            # Will add link to convert into lottie axis
            value = common.Vector.Vector()
            value[0] = (settings.lottie_format["w"]/2) / settings.PIX_PER_UNIT
            value[1] = -(settings.lottie_format["h"]/2) / settings.PIX_PER_UNIT
            self.add_convert_link(value)

        self.expression_controllers = []
        self.recur_animate(anim_type)

    def recur_animate(self, anim_type):
        """
        Internal private method for animating
        """
        if anim_type in {"vector", "group_layer_scale", "stretch_layer_scale", "circle_radius"}:
            self.dimension = 2

        # Check if we are dealing with convert methods
        if self.param.tag in settings.BONES or self.param[0].tag in settings.CONVERT_METHODS:
            self.extract_subparams()

            if self.param.tag == "bone":  # Careful about param[0] and param here

                # Get the animation of this origin
                bone_origin, eff_1 = self.subparams["origin"].recur_animate("vector")

                # Get the animation of angle
                prop_name = "angle"
                if self.is_group_child:
                    prop_name = "rotate_layer_angle"
                bone_angle, ang_eff = self.subparams["angle"].recur_animate(prop_name)

                # Get the local length scale of the bone
                lls, lls_eff = self.subparams["scalelx"].recur_animate("scalar_multiply")

                # Get the recursive length scale of the bone
                rls, rls_eff = self.subparams["scalex"].recur_animate("scalar_multiply")

                # Adding to expression controller
                self.expression_controllers.extend(eff_1)
                self.expression_controllers.extend(ang_eff)
                self.expression_controllers.extend(lls_eff)
                self.expression_controllers.extend(rls_eff)

                # Now adding the parent's effects in this bone
                guid = self.subparams["parent"][0].attrib["guid"]
                canvas = self.get_canvas()
                bone = canvas.get_bone(guid)

                # Store the previous state
                prev_state = bone.is_group_child
                if self.is_group_child:
                    bone.is_group_child = True

                par_origin, par_angle, par_lls, par_rls, par_eff = bone.recur_animate("vector")

                # Restore the previous state
                bone.is_group_child = prev_state

                if par_eff is not None: # checking it is not the bone_root
                    # Adding expression to the expression controller
                    self.expression_controllers.extend(par_eff)

                    # Modifying the bone_origin according to the parent's local scale
                    bone_origin = "mul(" + bone_origin + ", " + par_lls + ")"

                    # Modifying the bone_origin according to the parent's recursive scale
                    bone_origin = "mul(" + bone_origin + ", " + par_rls + ")"

                    # Forming the expression for new origin
                    a1 = "degreesToRadians({angle})"
                    a2 = "degreesToRadians(sum({angle}, 90))"
                    a1, a2 = a1.format(angle=par_angle), a2.format(angle=par_angle)

                    ret_x = "sum({par_origin}[0], mul({bone_origin}[0], Math.cos({a1})), mul({bone_origin}[1], Math.cos({a2})))"
                    ret_y = "sum({par_origin}[1], mul({bone_origin}[0], Math.sin({a1})), mul({bone_origin}[1], Math.sin({a2})))"
                    ret_origin = "[" + ret_x + ", " + ret_y + "]"
                    ret_origin = ret_origin.format(par_origin=par_origin, bone_origin=bone_origin, a1=a1, a2=a2)

                    # Forming the expression for new angle
                    ret_angle = "sum({par_angle}, {bone_angle})"
                    ret_angle = ret_angle.format(par_angle=par_angle, bone_angle=bone_angle)
                else:
                    ret_origin = bone_origin
                    ret_angle = bone_angle

                ############# REMOVE AFTER DEBUGGING rls
                rls = "1"
                return ret_origin, ret_angle, lls, rls, self.expression_controllers

            elif self.param.tag == "bone_root": # No animation to be added as this being the root
                return "[0, 0]", "0", "1", "1", None

            elif self.param[0].tag == "add":
                self.subparams["add"].extract_subparams()
                lhs, effects_1 = self.subparams["add"].subparams["lhs"].recur_animate(anim_type)
                rhs, effects_2 = self.subparams["add"].subparams["rhs"].recur_animate(anim_type)   # Only one of the child should be converted to lottie axis
                scalar, effects_3 = self.subparams["add"].subparams["scalar"].recur_animate("scalar_multiply")
                self.expression_controllers.extend(effects_1)
                self.expression_controllers.extend(effects_2)
                self.expression_controllers.extend(effects_3)
                ret = "mul(sum({lhs}, {rhs}), {scalar})"
                ret = ret.format(lhs=lhs, rhs=rhs, scalar=scalar)
                self.expression = ret
                return ret, self.expression_controllers # Might have to return copy.deepcopy because these expressions are destroyed on animating this parameter again, and this param might be child to 2 params

            elif self.param[0].tag == "exp":
                self.subparams["exp"].extract_subparams()
                exp, effects_1 = self.subparams["exp"].subparams["exp"].recur_animate("scalar_multiply")
                scale, effects_2 = self.subparams["exp"].subparams["scale"].recur_animate(anim_type)
                self.expression_controllers.extend(effects_1)
                self.expression_controllers.extend(effects_2)
                ret = "mul(Math.exp({exp}), {scale})"
                ret = ret.format(exp=exp, scale=scale)
                self.expression = ret
                return ret, self.expression_controllers

            elif self.param[0].tag == "average":
                self.subparams["average"].extract_subparams()   # Entries will be extracted here
                lst = self.subparams["average"].subparams["entry"]
                if not isinstance(lst, list):   # When only one entry is present
                    lst = [lst]

                ret = "sum("    # Returning string
                for it in lst:
                    ent, eff = it.recur_animate(anim_type) 
                    self.expression_controllers.extend(eff)
                    ret += ent
                    ret += ","
                ret = ret[:-1]
                ret += ")"

                # Dividing by the length
                ret = "div(" + ret + "," + str(len(lst)) + ")"
                self.expression = ret
                return ret, self.expression_controllers

            elif self.param[0].tag == "composite":  # composite only on vectors
                self.subparams["composite"].extract_subparams()
                x, effect_1 = self.subparams["composite"].subparams["x"].recur_animate("real")
                y, effect_2 = self.subparams["composite"].subparams["y"].recur_animate("real")
                self.expression_controllers.extend(effect_1)
                self.expression_controllers.extend(effect_2)
                ret = "[{x}, -{y}]"
                ret = ret.format(x=x, y=y)
                self.expression = ret
                return ret, self.expression_controllers

            elif self.param[0].tag == "linear":
                self.subparams["linear"].extract_subparams()
                slope, effect_1 = self.subparams["linear"].subparams["slope"].recur_animate(anim_type)
                offset, effect_2 = self.subparams["linear"].subparams["offset"].recur_animate(anim_type)
                self.expression_controllers.extend(effect_1)
                self.expression_controllers.extend(effect_2)
                ret = "sum({offset}, mul({slope}, time))"
                ret = ret.format(offset=offset, slope=slope)
                self.expression = ret
                return ret, self.expression_controllers

            elif self.param[0].tag == "radial_composite":   # only for vectors
                prop_name = "angle"
                sett = settings.GROUP_LAYER.copy()
                sett.update(settings.PRE_COMP_LAYER)
                if self.get_layer().get_type() in sett: # depending upon the layer type, we use the angle is positive/negative value
                    prop_name = "rotate_layer_angle"

                self.subparams["radial_composite"].extract_subparams()
                rad, eff_1 = self.subparams["radial_composite"].subparams["radius"].recur_animate("real")
                ang, eff_2 = self.subparams["radial_composite"].subparams["theta"].recur_animate(prop_name)
                self.expression_controllers.extend(eff_1)
                self.expression_controllers.extend(eff_2)
                ret = "[mul({radius}, Math.cos(degreesToRadians({angle}))), mul({radius}, Math.sin(degreesToRadians({angle})))]"
                ret = ret.format(radius=rad, angle=ang)
                self.expression = ret
                return ret, self.expression_controllers

            elif self.param[0].tag == "scale":
                self.subparams["scale"].extract_subparams()
                link, eff_1 = self.subparams["scale"].subparams["link"].recur_animate(anim_type)
                scalar, eff_2 = self.subparams["scale"].subparams["scalar"].recur_animate("scalar_multiply")
                self.expression_controllers.extend(eff_1)
                self.expression_controllers.extend(eff_2)
                if settings.RANGE_FLAG == 1:
                    ret = "{link}"
                    ret = ret.format(link=link)
                else:
                    ret = "mul({link}, {scalar})"
                    ret = ret.format(link=link, scalar=scalar)
                self.expression = ret
                return ret, self.expression_controllers

            elif self.param[0].tag == "subtract":
                self.subparams["subtract"].extract_subparams()
                scalar, eff_1 = self.subparams["subtract"].subparams["scalar"].recur_animate("scalar_multiply")
                lhs, eff_2 = self.subparams["subtract"].subparams["lhs"].recur_animate(anim_type)
                rhs, eff_3 = self.subparams["subtract"].subparams["rhs"].recur_animate(anim_type)
                self.expression_controllers.extend(eff_1)
                self.expression_controllers.extend(eff_2)
                self.expression_controllers.extend(eff_3)
                ret = "mul(sub({lhs}, {rhs}), {scalar})"
                ret = ret.format(lhs=lhs, rhs=rhs, scalar=scalar)
                self.expression = ret
                return ret, self.expression_controllers

            elif self.param[0].tag == "switch":
                self.subparams["switch"].extract_subparams()
                link_off, eff_1 = self.subparams["switch"].subparams["link_off"].recur_animate(anim_type)
                link_on, eff_2 = self.subparams["switch"].subparams["link_on"].recur_animate(anim_type)
                switch, eff_3 = self.subparams["switch"].subparams["switch"].recur_animate("bool")
                self.expression_controllers.extend(eff_1)
                self.expression_controllers.extend(eff_2)
                self.expression_controllers.extend(eff_3)
                ret = "sum(mul({eff_2}, {eff_3}), mul({eff_1}, sub(1, {eff_3})))"
                ret = ret.format(eff_1=link_off, eff_2=link_on, eff_3=switch)
                self.expression = ret
                return ret, self.expression_controllers

            elif self.param[0].tag == "weighted_average":
                self.subparams["weighted_average"].extract_subparams()
                lst = self.subparams["weighted_average"].subparams["entry"]
                if not isinstance(lst, list):   # When only one entry is present
                    lst = [lst]

                ret = "sum("    # Returning string
                den = "sum("    # Denominator string
                for it in lst:
                    it.extract_subparams()  # weighted_vector
                    it.subparams["weighted_vector"].extract_subparams()
                    weight, eff_1 = it.subparams["weighted_vector"].subparams["weight"].recur_animate("scalar_multiply")
                    value, eff_2 = it.subparams["weighted_vector"].subparams["value"].recur_animate(anim_type)
                    self.expression_controllers.extend(eff_1)
                    self.expression_controllers.extend(eff_2)
                    hell = "mul({weight}, {value}),"
                    hell = hell.format(weight=weight, value=value)
                    ret += hell

                    den += weight
                    den += ","
                ret = ret[:-1]
                den = den[:-1]
                den += ")"
                ret += ")"
                ret = "div(" + ret + "," + den + ")"
                self.expression = ret
                return ret, self.expression_controllers

            elif self.param[0].tag == "bone_link":
                self.subparams["bone_link"].extract_subparams()
                guid = self.subparams["bone_link"].subparams["bone"][0].attrib["guid"]
                bone = self.get_bone_from_canvas(guid)

                base_value, bv_eff = self.subparams["bone_link"].subparams["base_value"].recur_animate("vector")

                # Storing previous state
                prev_state = bone.is_group_child
                if self.is_group_child:
                    bone.is_group_child = True
                origin, angle, lls, rls, eff = bone.recur_animate("vector")

                # Adding lls effect to the bone
                base_value = "mul(" + base_value + ", " + lls + ")"

                # Adding rls effect to the bone
                base_value = "mul(" + base_value + ", " + rls + ")"

                # Adding effect of base_value here
                a1 = "degreesToRadians({angle})"
                a2 = "degreesToRadians(sum({angle}, 90))"
                a1, a2 = a1.format(angle=angle), a2.format(angle=angle)

                ret_x = "sub(sum({origin}[0], mul({base_value}[0], Math.cos({a1}))), mul({base_value}[1], Math.cos({a2})))"
                ret_y = "sub(sum({origin}[1], mul({base_value}[0], Math.sin({a1}))), mul({base_value}[1], Math.sin({a2})))"
                ret_origin = "[" + ret_x + ", -" + ret_y + "]"
                ret_origin = ret_origin.format(origin=origin, base_value=base_value, a1=a1, a2=a2)

                # Restore previous state
                bone.is_group_child = prev_state

                if eff is not None:
                    self.expression_controllers.extend(eff)
                self.expression_controllers.extend(bv_eff)

                self.expression = ret_origin
                return ret_origin, self.expression_controllers

            elif self.param[0].tag == "sine":
                self.subparams["sine"].extract_subparams()
                angle, eff_1 = self.subparams["sine"].subparams["angle"].recur_animate("region_angle")
                amp, eff_2 = self.subparams["sine"].subparams["amp"].recur_animate("real")
                self.expression_controllers.extend(eff_1)
                self.expression_controllers.extend(eff_2)
                
                if self.dimension == 2:
                    ret = "mul(Math.sin(degreesToRadians({angle})), [{amp}, {amp}])"
                else:
                    ret = "mul(Math.sin(degreesToRadians({angle})),{amp})"
                ret = ret.format(angle=angle,amp=amp)

                self.expression = ret
                return ret, self.expression_controllers

            elif self.param[0].tag == "cos":
                self.subparams["cos"].extract_subparams()
                angle, eff_1 = self.subparams["cos"].subparams["angle"].recur_animate("region_angle")
                amp, eff_2 = self.subparams["cos"].subparams["amp"].recur_animate("real")
                self.expression_controllers.extend(eff_1)
                self.expression_controllers.extend(eff_2)
                
                if self.dimension == 2:
                    ret = "mul(Math.cos(degreesToRadians({angle})), [{amp}, {amp}])"
                else:
                    ret = "mul(Math.cos(degreesToRadians({angle})),{amp})"
                ret = ret.format(angle=angle,amp=amp)

                self.expression = ret
                return ret, self.expression_controllers

            elif self.param[0].tag == "fromint":
                self.subparams["fromint"].extract_subparams()
                if self.dimension == 2:
                    link, eff_1 = self.subparams["fromint"].subparams["link"].recur_animate("scalar_multiply")
                    self.expression_controllers.extend(eff_1)
                    ret = "[mul(Math.round({link}), {PIX_PER_UNIT}), mul(Math.round({link}), {PIX_PER_UNIT})]"
                    ret = ret.format(link=link, PIX_PER_UNIT=settings.PIX_PER_UNIT)
                else:
                    link, eff_1 = self.subparams["fromint"].subparams["link"].recur_animate("scalar_multiply")
                    self.expression_controllers.extend(eff_1)
                    ret = "mul(Math.round({link}), {PIX_PER_UNIT})"
                    ret = ret.format(link=link, PIX_PER_UNIT=settings.PIX_PER_UNIT)
                self.expression = ret
                return ret, self.expression_controllers

            elif self.param[0].tag == "atan2":
                self.subparams["atan2"].extract_subparams()
                y, eff_1 = self.subparams["atan2"].subparams["y"].recur_animate(anim_type)
                x, eff_2 = self.subparams["atan2"].subparams["x"].recur_animate(anim_type)
                self.expression_controllers.extend(eff_1)
                self.expression_controllers.extend(eff_2)
                
                ret = "sub(180,radiansToDegrees(Math.atan2({y}, {x})))"
                ret = ret.format(y=y,x=x)
                self.expression = ret
                return ret, self.expression_controllers

            elif self.param[0].tag == "vectorangle":
                self.subparams["vectorangle"].extract_subparams()
                vector, eff_1 = self.subparams["vectorangle"].subparams["vector"].recur_animate("vector")
                self.expression_controllers.extend(eff_1)
                ret = "radiansToDegrees(Math.atan2({y}[1], {x}[0]))"
                ret = ret.format(y=vector,x=vector)

                self.expression = ret
                return ret, self.expression_controllers

            elif self.param[0].tag == "power":
                self.subparams["power"].extract_subparams()
                base, eff_1     = self.subparams["power"].subparams["base"].recur_animate("scalar_multiply")
                power, eff_2    = self.subparams["power"].subparams["power"].recur_animate("scalar_multiply")
                epsilon, eff_3  = self.subparams["power"].subparams["epsilon"].recur_animate("scalar_multiply")
                infinite, eff_4 = self.subparams["power"].subparams["infinite"].recur_animate("scalar_multiply")

                self.expression_controllers.extend(eff_1)
                self.expression_controllers.extend(eff_2)
                self.expression_controllers.extend(eff_3)
                self.expression_controllers.extend(eff_4)

                if self.dimension == 2:
                    ret = "[mul(Math.pow({base}, {power}),{PIX_PER_UNIT}),mul(Math.pow({base}, {power}),{PIX_PER_UNIT})]"
                else:
                    ret = "mul(Math.pow({base}, {power}),{PIX_PER_UNIT})"

                ret = ret.format(base=base,power=power,PIX_PER_UNIT=settings.PIX_PER_UNIT)
                self.expression = ret
                return ret, self.expression_controllers

            elif self.param[0].tag == "vectorx":
                self.subparams["vectorx"].extract_subparams()
                vector, eff_1 = self.subparams["vectorx"].subparams["vector"].recur_animate("vector")
                self.expression_controllers.extend(eff_1)
                ret = "[{x}[0], {x}[0]]" if self.dimension == 2 else "{x}[0]"
                ret = ret.format(x=vector)
                self.expression = ret
                return ret, self.expression_controllers

            elif self.param[0].tag == "vectory":
                self.subparams["vectory"].extract_subparams()
                vector, eff_1 = self.subparams["vectory"].subparams["vector"].recur_animate("vector")
                self.expression_controllers.extend(eff_1)
                ret = "[{y}[1], {y}[1]]" if self.dimension == 2 else "{y}[1]"
                ret = ret.format(y=vector)
                self.expression = ret
                return ret, self.expression_controllers

            elif self.param[0].tag == "dotproduct":
                self.subparams["dotproduct"].extract_subparams()
                lhs, eff_1 = self.subparams["dotproduct"].subparams["lhs"].recur_animate("vector")
                rhs, eff_2 = self.subparams["dotproduct"].subparams["rhs"].recur_animate("vector")

                self.expression_controllers.extend(eff_1)
                self.expression_controllers.extend(eff_2)

                if anim_type == 'real':
                    settings.DOT_FLAG = 0
                    x_comp = "mul(div({lhs}[0],{PIX_PER_UNIT}),div({rhs}[0],{PIX_PER_UNIT}))"
                    y_comp = "mul(div({lhs}[1],{PIX_PER_UNIT}),div({rhs}[1],{PIX_PER_UNIT}))"

                    x_comp = x_comp.format(lhs=lhs,rhs=rhs,PIX_PER_UNIT=settings.PIX_PER_UNIT)
                    y_comp = y_comp.format(lhs=lhs,rhs=rhs,PIX_PER_UNIT=settings.PIX_PER_UNIT)

                    if self.dimension == 2:
                        ret = "[mul(sum({x_comp},{y_comp}),{PIX_PER_UNIT}),mul(sum({x_comp},{y_comp}),{PIX_PER_UNIT})]"
                    else:
                        ret = "mul(sum({x_comp},{y_comp}),{PIX_PER_UNIT})"
                    ret = ret.format(x_comp=x_comp,y_comp=y_comp,PIX_PER_UNIT=settings.PIX_PER_UNIT)

                else:
                    settings.DOT_FLAG = 1
                    x_comp = "mul(div({lhs}[0],{PIX_PER_UNIT}),div({rhs}[0],{PIX_PER_UNIT}))"
                    y_comp = "mul(div({lhs}[1],{PIX_PER_UNIT}),div({rhs}[1],{PIX_PER_UNIT}))"

                    x_comp = x_comp.format(lhs=lhs,rhs=rhs,PIX_PER_UNIT=settings.PIX_PER_UNIT)
                    y_comp = y_comp.format(lhs=lhs,rhs=rhs,PIX_PER_UNIT=settings.PIX_PER_UNIT)

                    vector_magnitude_1 = 'Math.sqrt(sum(Math.pow(div({lhs}[0],{PIX_PER_UNIT}),2),Math.pow(div({lhs}[1],{PIX_PER_UNIT}),2)))'
                    vector_magnitude_1 = vector_magnitude_1.format(lhs=lhs,PIX_PER_UNIT=settings.PIX_PER_UNIT)

                    vector_magnitude_2 = 'Math.sqrt(sum(Math.pow(div({rhs}[0],{PIX_PER_UNIT}),2),Math.pow(div({rhs}[1],{PIX_PER_UNIT}),2)))'
                    vector_magnitude_2 = vector_magnitude_2.format(rhs=rhs,PIX_PER_UNIT=settings.PIX_PER_UNIT)

                    mult = "mul({vector_magnitude_1},{vector_magnitude_2})"
                    mult = mult.format(vector_magnitude_1=vector_magnitude_1,vector_magnitude_2=vector_magnitude_2)

                    ret = "Math.acos(div(sum({x_comp},{y_comp})),{mult}))"
                    ret = ret.format(x_comp=x_comp,y_comp=y_comp,mult=mult)

                self.expression = ret
                return ret, self.expression_controllers

            elif self.param[0].tag == "reciprocal":
                self.subparams["reciprocal"].extract_subparams()
                link, eff_1 = self.subparams["reciprocal"].subparams["link"].recur_animate("scalar_multiply")
                epsilon, eff_2 = self.subparams["reciprocal"].subparams["epsilon"].recur_animate("scalar_multiply")
                infinite, eff_3 = self.subparams["reciprocal"].subparams["infinite"].recur_animate("scalar_multiply")
                
                self.expression_controllers.extend(eff_1)
                self.expression_controllers.extend(eff_2)
                self.expression_controllers.extend(eff_3)
                ret = "if({link} <= -{epsilon} || {epsilon} <= {link})\n $bm_rt = [mul(div({PIX_PER_UNIT},{link}),2),mul(div({PIX_PER_UNIT},{link}),2)] ;\n else if({link} >= 0 && {epsilon} > {link})\n $bm_rt = [{infinite},{infinite}];\n else\n $bm_rt = [-{infinite},-{infinite}];"
                ret = ret.format(link=link,epsilon=epsilon,infinite=infinite,PIX_PER_UNIT=settings.PIX_PER_UNIT)
                
            elif self.param[0].tag == "range":
                self.subparams["range"].extract_subparams()
                min_val, eff_1 = self.subparams["range"].subparams["min"].recur_animate("real")
                max_val, eff_2 = self.subparams["range"].subparams["max"].recur_animate("real")
                link   , eff_3 = self.subparams["range"].subparams["link"].recur_animate("real")

                self.expression_controllers.extend(eff_1)
                self.expression_controllers.extend(eff_2)
                self.expression_controllers.extend(eff_3)

                ret = "if({link} < {min_val})\n $bm_rt = [mul({min_val},2),mul({min_val},2)] ;\n else if({link} > {max_val})\n $bm_rt = [mul({max_val},2),mul({max_val},2)];\n else\n $bm_rt = [mul({link},2),mul({link},2)];"
                ret = ret.format(link=link,min_val=min_val,max_val=max_val)
                settings.RANGE_FLAG = 1
                self.expression = ret
                return ret,self.expression_controllers

            elif self.param[0].tag == "logarithm":
                self.subparams["logarithm"].extract_subparams()
                link, eff_1     = self.subparams["logarithm"].subparams["link"].recur_animate("real")
                epsilon, eff_2    = self.subparams["logarithm"].subparams["epsilon"].recur_animate("real")
                infinite, eff_3  = self.subparams["logarithm"].subparams["infinite"].recur_animate("real")
                
                self.expression_controllers.extend(eff_1)
                self.expression_controllers.extend(eff_2)
                self.expression_controllers.extend(eff_3)
                
                if self.dimension == 2:
                    ret = "[mul(Math.log(div({link},{PIX_PER_UNIT})),{PIX_PER_UNIT}),mul(Math.log(div({link},{PIX_PER_UNIT})),{PIX_PER_UNIT})]"
                else:
                    ret = "mul(Math.log(div({link},{PIX_PER_UNIT})),{PIX_PER_UNIT})"

                ret = ret.format(link=link,PIX_PER_UNIT=settings.PIX_PER_UNIT)
                self.expression = ret
                return ret, self.expression_controllers
            
            elif self.param[0].tag == "vectorlength":
                self.subparams["vectorlength"].extract_subparams()
                vector, eff_1 = self.subparams["vectorlength"].subparams["vector"].recur_animate("vector")
                self.expression_controllers.extend(eff_1)
                ret = "[Math.sqrt(sum(Math.pow({vector}[1],2), Math.pow({vector}[0],2))),Math.sqrt(sum(Math.pow({vector}[1],2), Math.pow({vector}[0],2)))]" if self.dimension == 2 else "Math.sqrt(sum(Math.pow({vector}[1],2), Math.pow({vector}[0],2)))"
                ret = ret.format(vector=vector)
                self.expression = ret
                return ret, self.expression_controllers

        else:
            self.single_animate(anim_type)
            # Insert the animation into the effect
            self.expression_controllers.append({})

            if anim_type == "vector":
                self.gen_path("vector")
            elif anim_type not in {"bool", "string"}:   # These do not need path generation
                self.gen_path("real")   # "real" can be of various types as defined in common.misc.parse_position()
            elif anim_type in {"bool"}:     # This is the first time bool is needed
                self.gen_path("bool")

            
            gen_effects_controller(self.expression_controllers[-1], self.get_path(), anim_type)

            # Extract name from the effects
            ret = "effect('{effect_1}')('{effect_2}')"
            ret = ret.format(effect_1=self.expression_controllers[-1]["nm"], effect_2=self.expression_controllers[-1]["ef"][0]["nm"])
            self.expression = ret
            return ret, self.expression_controllers

    def single_animate(self, anim_type):
        """
        For animating a single animation, convert methods will use this internal
        function
        """
        is_animate = common.misc.is_animated(self.param[0])
        if is_animate == settings.ANIMATED:
            # If already animated, no need to add waypoints
            # Forcibly set it's animation type to the anim_type
            self.param[0].attrib["type"] = anim_type
            if anim_type == "bool":
                modify_bool_animation(self.param[0])
            return
        elif is_animate == settings.NOT_ANIMATED:
            st = '<animated type="{anim_type}"><waypoint time="0s" before="constant" after="constant"></waypoint></animated>'
            st = st.format(anim_type=anim_type)
            root = etree.fromstring(st)
            root[0].append(copy.deepcopy(self.param[0]))
            self.param[0] = root
        elif is_animate == settings.SINGLE_WAYPOINT:
            self.param[0].attrib["type"] = anim_type
            self.param[0][0].attrib["before"] = self.param[0][0].attrib["after"] = "constant"

        new_waypoint = copy.deepcopy(self.param[0][0])
        frame = common.misc.get_frame(self.param[0][0])
        frame += 1
        time = frame / settings.lottie_format["fr"]
        time = str(time) + 's'
        new_waypoint.attrib["time"] = time
        self.param[0].insert(1, new_waypoint)

        if anim_type == "bool":
            modify_bool_animation(self.param[0])

    def gen_path(self, anim_type="real", idx=0):
        """
        Generates the path for this parameter over time depending on the
        animation type of this parameter
        """
        self.path = {}
        if anim_type in {"real", "bool"}:
            gen_value_Keyframed(self.path, self.param[0], idx)
        else:
            gen_properties_multi_dimensional_keyframed(self.path, self.param[0], idx)

    def get_path(self):
        """
        Returns the dictionary which stores the path of this parameter
        """
        return self.path

    def fill_path(self, lottie, key):
        """
        Fill's the lottie dictionary with the path of the parameter
        """
        if self.param[0].tag in settings.CONVERT_METHODS:
            if settings.RANGE_FLAG == 1:
                expression = "var $bm_rt; {expr}"
                settings.RANGE_FLAG = 0
            else:
                expression = "var $bm_rt; $bm_rt = {expr}"
            expression = expression.format(expr=self.expression)
            if self.dimension == 1:
                val = 1
            else:
                val = [1, 1]
            gen_properties_value(lottie[key],
                                 val,
                                 0,
                                 0,
                                 expression)
            if "ef" not in self.get_layer().get_lottie_layer().keys():
                self.get_layer().get_lottie_layer()["ef"] = []
            self.get_layer().get_lottie_layer()["ef"].extend(self.expression_controllers)
                
        else:
            lottie[key] = copy.deepcopy(self.path)

    def get_value(self, frame):
        """
        Public method to get the value of the parameter at a given frame
        """
        ret = self.__get_value(frame)
        # Convert into Lottie format
        if isinstance(ret, list):
            ret = [ret[0], -ret[1]]
        return ret
    
    def __get_value(self, frame):
        """
        Returns the value of the parameter at a given frame
        """
        if self.param.tag in settings.BONES or self.param[0].tag in settings.CONVERT_METHODS:
            if self.param.tag == "bone":
                cur_origin = self.subparams["origin"].__get_value(frame)

                # Now adding the parent's effects in this bone
                guid = self.subparams["parent"][0].attrib["guid"]
                canvas = self.get_canvas()
                bone = canvas.get_bone(guid)
                shifted_origin, shifted_angle, lls, rls = bone.__get_value(frame)
                a1, a2 = math.radians(shifted_angle), math.radians(shifted_angle+90)

                # Calculating this bones angle with respect to parent bone's
                # angle
                if "angle" in self.subparams.keys():
                    angle = to_Synfig_axis(self.subparams["angle"].__get_value(frame), "angle")
                else:
                    angle = 0

                # Calculating the local length scale
                local_length_scale = self.subparams["scalelx"].__get_value(frame)

                # Calculating the recursive length scale
                this_rls = self.subparams["scalex"].__get_value(frame)    # In current angle's direction
                absolute_angle = shifted_angle+angle
                aa1 = math.radians(absolute_angle)
                this_rls = [this_rls * math.cos(aa1), this_rls * math.sin(aa1)]

                # Calculate returning recursive length
                ret_rls = [this_rls[0]*rls[0], this_rls[1]*rls[1]]
                ##### REMOVE AFTER DEBUGGING
                ret_rls = [1, 1]

                # Multiplying the current bone origin with the scale
                cur_origin = [i*lls for i in cur_origin]

                ret = shifted_origin
                # Adding effect of x component
                ret[0] = ret[0] + (cur_origin[0] * math.cos(a1) + cur_origin[1] * math.cos(a2)) * rls[0]
                ret[1] = ret[1] + (cur_origin[0] * math.sin(a1) + cur_origin[1] * math.sin(a2)) * rls[1]

                return ret, absolute_angle, local_length_scale, ret_rls

            elif self.param.tag == "bone_root":
                origin = [0, 0]
                angle = 0
                local_length_scale = 1
                recursive_length_scale = [1, 1] # x and y axis
                return origin, angle, local_length_scale, recursive_length_scale

            elif self.param[0].tag == "add":
                ret = self.subparams["add"].subparams["lhs"].__get_value(frame)
                ret2 = self.subparams["add"].subparams["rhs"].__get_value(frame)
                mul = self.subparams["add"].subparams["scalar"].__get_value(frame)
                if isinstance(ret, list):
                    ret[0] += ret2[0]
                    ret[1] += ret2[1]
                    ret = [it*mul for it in ret]
                else:
                    ret += ret2
                    ret *= mul

            elif self.param[0].tag == "exp":
                exp = self.subparams["exp"].subparams["exp"].__get_value(frame)
                scale = self.subparams["exp"].subparams["scale"].__get_value(frame)
                ret = scale * math.exp(exp)

            elif self.param[0].tag == "average":
                lst = self.subparams["average"].subparams["entry"]
                if not isinstance(lst, list):
                    lst = [lst]

                ret = [0, 0]
                if not isinstance(lst[0].__get_value(frame), list):
                    ret = 0
                for it in lst:
                    val = it.__get_value(frame)
                    if isinstance(val, list):
                        ret[0], ret[1] = ret[0] + val[0], ret[1] + val[1]
                    else:
                        ret += val
                if isinstance(ret, list):
                    ret[0], ret[1] = ret[0] / len(lst), ret[1] / len(lst)
                else:
                    ret /= float(len(lst))

            elif self.param[0].tag == "weighted_average":
                self.subparams["weighted_average"].extract_subparams()
                lst = self.subparams["weighted_average"].subparams["entry"]
                if not isinstance(lst, list):   # When only one entry is present
                    lst = [lst]

                ret = [0, 0]
                den = 0
                if not isinstance(lst[0].subparams["weighted_vector"].subparams["value"].__get_value(frame), list):
                    ret = 0
                for it in lst:
                    weight = it.subparams["weighted_vector"].subparams["weight"].__get_value(frame)
                    value = it.subparams["weighted_vector"].subparams["value"].__get_value(frame)
                    den += weight
                    if isinstance(value, list):
                        ret[0], ret[1] = ret[0] + value[0]*weight, ret[1] + value[1]*weight
                    else:
                        ret += value*weight
                if isinstance(ret, list):
                    ret[0], ret[1] = ret[0] / den, ret[1] / den
                else:
                    ret /= float(den)

            elif self.param[0].tag == "composite":  # Only available for vectors
                x = self.subparams["composite"].subparams["x"].__get_value(frame)
                y = self.subparams["composite"].subparams["y"].__get_value(frame)
                ret = [x, y]

            elif self.param[0].tag == "linear":
                slope = self.subparams["linear"].subparams["slope"].__get_value(frame)
                offset = self.subparams["linear"].subparams["offset"].__get_value(frame)
                if isinstance(slope, list):
                    ret = [0, 0]
                    ret[0] = offset[0] + slope[0]*(frame/settings.lottie_format["fr"])
                    ret[1] = offset[1] + slope[1]*(frame/settings.lottie_format["fr"])
                else:
                    ret = offset + slope*(frame/settings.lottie_format["fr"])

            elif self.param[0].tag == "radial_composite":   # Only for vectors
                rad = self.subparams["radial_composite"].subparams["radius"].__get_value(frame)
                angle = to_Synfig_axis(self.subparams["radial_composite"].subparams["theta"].__get_value(frame), "angle")
                angle = math.radians(angle)
                x = rad * math.cos(angle)
                y = rad * math.sin(angle)
                ret = [x, y]

            elif self.param[0].tag == "scale":
                link = self.subparams["scale"].subparams["link"].__get_value(frame)
                scalar = self.subparams["scale"].subparams["scalar"].__get_value(frame)
                if isinstance(link, list):
                    link[0] *= scalar
                    link[1] *= scalar
                else:
                    link *= scalar
                ret = link

            elif self.param[0].tag == "subtract":
                lhs = self.subparams["subtract"].subparams["lhs"].__get_value(frame)
                rhs = self.subparams["subtract"].subparams["rhs"].__get_value(frame)
                scalar = self.subparams["subtract"].subparams["scalar"].__get_value(frame)
                if isinstance(lhs, list):
                    ret = [0, 0]
                    ret[0] = (lhs[0] - rhs[0]) * scalar
                    ret[1] = (lhs[1] - rhs[1]) * scalar
                else:
                    ret = (lhs - rhs) * scalar

            elif self.param[0].tag == "switch":
                link_off = self.subparams["switch"].subparams["link_off"].__get_value(frame)
                link_on = self.subparams["switch"].subparams["link_on"].__get_value(frame)
                switch = self.subparams["switch"].subparams["switch"].__get_value(frame)
                if isinstance(link_on, list):
                    ret = [0, 0]
                    ret[0] = link_on[0] * switch + link_off[0] * (1 - switch)
                    ret[1] = link_on[1] * switch + link_off[1] * (1 - switch)
                else:
                    ret = link_on * switch + link_off * (1 - switch)

            elif self.param[0].tag == "bone_link":
                guid = self.subparams["bone_link"].subparams["bone"][0].attrib["guid"]
                bone = self.get_bone_from_canvas(guid)
                ret_origin, ret_angle, lls, rls = bone.__get_value(frame)

                # Adding the base value effect here
                base_value = self.subparams["bone_link"].subparams["base_value"].__get_value(frame)
                a1, a2 = math.radians(ret_angle), math.radians(ret_angle+90)
                ret = ret_origin

                # base_value to be arranged according to the local scale
                base_value = [lls*i for i in base_value]

                ret[0] = ret[0] + (base_value[0] * math.cos(a1) - base_value[1] * math.cos(a2)) * rls[0]
                ret[1] = ret[1] + (base_value[0] * math.sin(a1) - base_value[1] * math.sin(a2)) * rls[1]

                ret = [ret[0], ret[1]]

            elif self.param[0].tag == "sine":
                angle = self.subparams["sine"].subparams["angle"].__get_value(frame)
                amp = self.subparams["sine"].subparams["amp"].__get_value(frame)
                angle = math.radians(angle)
                
                if isinstance(amp, list):
                    ret = [0, 0]

                    ret[0] = math.sin(angle) * amp[0]
                    ret[1] = math.sin(angle) * amp[1]
                else:
                    ret = math.sin(angle)*amp

            elif self.param[0].tag == "cos":
                angle = self.subparams["cos"].subparams["angle"].__get_value(frame)
                amp = self.subparams["cos"].subparams["amp"].__get_value(frame)
                angle = math.radians(angle)
                if isinstance(amp, list):
                    ret = [0, 0]
                    ret[0] = math.cos(angle) * amp[0]
                    ret[1] = math.cos(angle) * amp[1]
                else:
                    ret = math.cos(angle)*amp

            elif self.param[0].tag == "fromint":
                link = self.subparams["fromint"].subparams["link"].__get_value(frame)
                if isinstance(link, list):
                    ret = [0, 0]
                    ret[0] = round(link[0])*settings.PIX_PER_UNIT
                    ret[1] = round(link[1])*settings.PIX_PER_UNIT
                else:
                    ret = round(link)*settings.PIX_PER_UNIT

            elif self.param[0].tag == "atan2":
                y = self.subparams["atan2"].subparams["y"].__get_value(frame)
                x = self.subparams["atan2"].subparams["x"].__get_value(frame)
                rad = math.pi/180
                ret = math.atan2(y,x)/rad

            elif self.param[0].tag == "vectorangle":
                vector = self.subparams["vectorangle"].subparams["vector"].__get_value(frame)
                rad = math.pi/180
                ret = math.atan2(vector[1],vector[0])/rad

            elif self.param[0].tag == "power":
                base = self.subparams["power"].subparams["base"].__get_value(frame)
                power = self.subparams["power"].subparams["power"].__get_value(frame)
                epsilon = self.subparams["power"].subparams["epsilon"].__get_value(frame)
                infinite = self.subparams["power"].subparams["infinite"].__get_value(frame)
                if epsilon < 0.00000001:
                    epsilon = 0.00000001

                #Filters for special/undefined cases
                if abs(power) < epsilon: #x^0 = 1
                    return 1*settings.PIX_PER_UNIT
                if abs(base) < epsilon:
                    if power > 0: #0^x=0
                        return 0
                    else:
                        if int(power) % 2 != 0 and base < 0: #(-0)^(-odd)=-inf
                            return -infinite
                        else:
                            return infinite

                if base <= epsilon and int(power) != power: #negative number to fractional power -> undefined
                    power = int(power)  #so round off power to nearest integer

                ret = math.pow(base,power)*settings.PIX_PER_UNIT

            elif self.param[0].tag == "vectorx":
                vector = self.subparams["vectorx"].subparams["vector"].__get_value(frame)
                ret = vector[0]

            elif self.param[0].tag == "vectory":
                vector = self.subparams["vectory"].subparams["vector"].__get_value(frame)
                ret = vector[1]

            elif self.param[0].tag == "dotproduct":
                lhs = self.subparams["dotproduct"].subparams["lhs"].__get_value(frame)
                rhs = self.subparams["dotproduct"].subparams["rhs"].__get_value(frame)
                dot = (lhs[0]*rhs[0]+lhs[1]*rhs[1])/(settings.PIX_PER_UNIT*settings.PIX_PER_UNIT)
                
                if settings.DOT_FLAG != 1:
                    ret = dot*settings.PIX_PER_UNIT
                else:
                    vector_magnitude_1 = math.sqrt(math.pow(lhs[0]/settings.PIX_PER_UNIT,2)+math.pow(lhs[1]/settings.PIX_PER_UNIT,2))
                    vector_magnitude_2 = math.sqrt(math.pow(rhs[0]/settings.PIX_PER_UNIT,2)+math.pow(rhs[1]/settings.PIX_PER_UNIT,2))
                    mult = vector_magnitude_1*vector_magnitude_2
                    ret = math.degrees(math.acos(dot/mult))

            elif self.param[0].tag == "reciprocal":
                link = self.subparams["reciprocal"].subparams["link"].__get_value(frame)
                epsilon = self.subparams["reciprocal"].subparams["epsilon"].__get_value(frame)
                infinite = self.subparams["reciprocal"].subparams["infinite"].__get_value(frame)

                if link <= -epsilon or epsilon <= link:
                    ret = settings.PIX_PER_UNIT/link

                elif link >= 0:
                    ret = infinite

                else:
                    ret = -infinite

            elif self.param[0].tag == "logarithm":
                link     = self.subparams["logarithm"].subparams["link"].__get_value(frame)
                epsilon  = self.subparams["logarithm"].subparams["epsilon"].__get_value(frame)
                infinite = self.subparams["logarithm"].subparams["infinite"].__get_value(frame)

                if link >= epsilon:
                    ret = settings.PIX_PER_UNIT*math.log(link/settings.PIX_PER_UNIT)
                else:
                    ret = -settings.PIX_PER_UNIT*infinite

            elif self.param[0].tag == "range":
                min_val = self.subparams["range"].subparams["min"].__get_value(frame)
                max_val = self.subparams["range"].subparams["max"].__get_value(frame)
                link    = self.subparams["range"].subparams["link"].__get_value(frame)

                if link < min_val:
                    ret = min_val
                elif link > max_val:
                    ret = max_val
                else:
                    ret = link

            elif self.param[0].tag == "vectorlength":
                vector = self.subparams["vectorlength"].subparams["vector"].__get_value(frame)
                ret = math.sqrt(math.pow(vector[1],2)+math.pow(vector[0],2))

        else:
            ret = self.get_single_value(frame)
            if isinstance(ret, list):
                # Need to change the calculation inside get_single_value, this
                # is just a hack
                ret = [ret[0], -ret[1]]
        return ret

    def get_single_value(self, frame):
        """
        Returns the value of some parameter which is not a convert method
        """
        if self.param[0].attrib["type"] == "bool":  # No need of lottie format path here
            return get_bool_at_frame(self.param[0], frame)

        if not self.path:   # Empty dictionary
            raise KeyError("Please calculate the path of this parameter before getting value at a frame")
        return get_vector_at_frame(self.path, frame)


    def add_offset(self):
        """
        Will modify the animation by inserting <add></add> xml elements
        """
        offset = synfig.group.get_offset()
        self.add_convert_link(offset)


    def add_convert_link(self, val):
        """
        Private method for inserting <add></add>, given a offset
        """
        st = "<add type='vector'><lhs></lhs><rhs></rhs><scalar><real value='1.00'/></scalar></add>"
        root = etree.fromstring(st)
        first = copy.deepcopy(self.param[0])
        root[0].append(first)
        
        st = "<vector><x>{x_val}</x><y>{y_val}</y></vector>"
        st = st.format(x_val=val[0], y_val=val[1])
        second = etree.fromstring(st)
        root[1].append(second)

        self.param[0].getparent().remove(self.param[0])
        self.param.append(root)
        self.SUBPARAMS_EXTRACTED = 0

    def scale_convert_link(self, val):
        """
        Private method for inserting <scale></scale>, given the multiplication value
        """
        st = "<scale type='real'><link></link><scalar><real value='1.00'/></scalar></scale>"
        root = etree.fromstring(st)
        first = copy.deepcopy(self.param[0])
        root[0].append(first)
        root[1][0].attrib["value"] = str(val)
        self.param[0].getparent().remove(self.param[0])
        self.param.append(root)
        self.SUBPARAMS_EXTRACTED = 0

    def update_frame_window(self, window):
        """
        Given an animation, find the minimum and maximum frame at which the
        waypoints are located
        """
        self.extract_subparams()
        if self.param.tag in {"bone", "bone_root"}:
            node = self.param
        else:
            node = self.param[0]

        if node.tag in settings.BONES:
            if node.tag == "bone":
                self.subparams["origin"].update_frame_window(window)
                self.subparams["angle"].update_frame_window(window)
                self.subparams["scalelx"].update_frame_window(window)
                parent_guid = self.subparams["parent"][0].attrib["guid"]
                canvas = self.get_canvas()
                bone = canvas.get_bone(parent_guid)
                bone.update_frame_window(window)

        # Time updation for converted nodes:
        if node.tag in settings.CONVERT_METHODS:
            if node.tag in {"add", "subtract"}:
                key = node.tag
                self.subparams[key].extract_subparams()
                self.subparams[key].subparams["lhs"].update_frame_window(window)
                self.subparams[key].subparams["rhs"].update_frame_window(window)
                self.subparams[key].subparams["scalar"].update_frame_window(window)

            elif node.tag == "switch":
                key = node.tag
                self.subparams[key].extract_subparams()
                self.subparams[key].subparams["link_off"].update_frame_window(window)
                self.subparams[key].subparams["link_on"].update_frame_window(window)
                self.subparams[key].subparams["switch"].update_frame_window(window)

            elif node.tag == "average":
                if not isinstance(self.subparams["average"], list):
                    self.subparams["average"] = [self.subparams["average"]]
                for it in self.subparams["average"]:
                    it.update_frame_window(window)

            elif node.tag == "exp":
                self.subparams["exp"].extract_subparams()
                self.subparams["exp"].subparams["exp"].update_frame_window(window)
                self.subparams["exp"].subparams["scale"].update_frame_window(window)

            elif node.tag == "weighted_average":
                self.subparams["weighted_average"].extract_subparams()
                for it in self.subparams["weighted_average"].subparams["entry"]:
                    it.extract_subparams()
                    it.subparams["weighted_vector"].extract_subparams()
                    ti = it.subparams["weighted_vector"].subparams
                    ti["weight"].update_frame_window(window)
                    ti["value"].update_frame_window(window)

            elif node.tag == "composite":
                self.subparams["composite"].extract_subparams()
                self.subparams["composite"].subparams["x"].update_frame_window(window)
                self.subparams["composite"].subparams["y"].update_frame_window(window)

            elif node.tag == "linear":
                window["first"] = settings.lottie_format["ip"]
                window["last"] = settings.lottie_format["op"]

            elif node.tag == "radial_composite":
                self.subparams["radial_composite"].extract_subparams()
                self.subparams["radial_composite"].subparams["radius"].update_frame_window(window)
                self.subparams["radial_composite"].subparams["theta"].update_frame_window(window)

            elif node.tag == "scale":
                self.subparams["scale"].extract_subparams()
                self.subparams["scale"].subparams["link"].update_frame_window(window)
                self.subparams["scale"].subparams["scalar"].update_frame_window(window)

            elif node.tag == "bone_link":
                self.subparams["bone_link"].extract_subparams()
                guid = self.subparams["bone_link"].subparams["bone"][0].attrib["guid"]
                bone = self.get_bone_from_canvas(guid)
                bone.update_frame_window(window)
                self.subparams["bone_link"].subparams["base_value"].update_frame_window(window)

            elif node.tag == "sine":
                self.subparams["sine"].extract_subparams()
                self.subparams["sine"].subparams["angle"].update_frame_window(window)
                self.subparams["sine"].subparams["amp"].update_frame_window(window)

            elif node.tag == "cos":
                self.subparams["cos"].extract_subparams()
                self.subparams["cos"].subparams["angle"].update_frame_window(window)
                self.subparams["cos"].subparams["amp"].update_frame_window(window)

            elif node.tag == "fromint":
                self.subparams["fromint"].extract_subparams()
                self.subparams["fromint"].subparams["link"].update_frame_window(window)

            elif node.tag == "atan2":
                self.subparams["atan2"].extract_subparams()
                self.subparams["atan2"].subparams["y"].update_frame_window(window)
                self.subparams["atan2"].subparams["x"].update_frame_window(window)

            elif node.tag == "vectorangle":
                self.subparams["vectorangle"].extract_subparams()
                self.subparams["vectorangle"].subparams["vector"].update_frame_window(window)

            elif node.tag == "power":
                self.subparams["power"].extract_subparams()
                self.subparams["power"].subparams["base"].update_frame_window(window)
                self.subparams["power"].subparams["power"].update_frame_window(window)
                self.subparams["power"].subparams["epsilon"].update_frame_window(window)
                self.subparams["power"].subparams["infinite"].update_frame_window(window)

            elif node.tag == "vectorx":
                self.subparams["vectorx"].extract_subparams()
                self.subparams["vectorx"].subparams["vector"].update_frame_window(window)

            elif node.tag == "vectory":
                self.subparams["vectory"].extract_subparams()
                self.subparams["vectory"].subparams["vector"].update_frame_window(window)

            elif node.tag == "dotproduct":
                self.subparams["dotproduct"].extract_subparams()
                self.subparams["dotproduct"].subparams["lhs"].update_frame_window(window)
                self.subparams["dotproduct"].subparams["rhs"].update_frame_window(window)

            elif node.tag == "reciprocal":
                self.subparams["reciprocal"].extract_subparams()
                self.subparams["reciprocal"].subparams["link"].update_frame_window(window)
                self.subparams["reciprocal"].subparams["epsilon"].update_frame_window(window)
                self.subparams["reciprocal"].subparams["infinite"].update_frame_window(window)

            elif node.tag == "logarithm":
                self.subparams["logarithm"].extract_subparams()
                self.subparams["logarithm"].subparams["link"].update_frame_window(window)
                self.subparams["logarithm"].subparams["epsilon"].update_frame_window(window)
                self.subparams["logarithm"].subparams["infinite"].update_frame_window(window)

            elif node.tag == "range":
                self.subparams["range"].extract_subparams()
                self.subparams["range"].subparams["min"].update_frame_window(window)
                self.subparams["range"].subparams["max"].update_frame_window(window)
                self.subparams["range"].subparams["link"].update_frame_window(window)

            elif node.tag == "vectorlength":
                self.subparams["vectorlength"].extract_subparams()
                self.subparams["vectorlength"].subparams["vector"].update_frame_window(window)

        if is_animated(node) == settings.ANIMATED:
            for waypoint in node:
                fr = common.misc.get_frame(waypoint)
                settings.WAYPOINTS_LIST.append(fr)
                if fr > window["last"]:
                    window["last"] = fr
                if fr < window["first"]:
                    window["first"] = fr

    def get_bone_from_canvas(self, guid):
        """
        Given a canvas, recursively travel to the ancestor canvas's and find the
        bone associated with the guid
        """
        now = self
        while True:
            layer = now.get_layer()
            canvas = layer.getparent()
            bone = canvas.get_bone(guid)
            if bone is not None:
                break
            else:
                now = canvas.getparent_param()
        return bone



"""
Count.py
Will store the Count class
"""

import sys
sys.path.append("..")


class Count:
    """
    Class to keep count of variable
    """
    def __init__(self):
        """
        Args:
            (None)

        Returns:
            (None)
        """
        self.idx = -1

    def inc(self):
        """
        This method increases the count by 1 and returns the new count

        Args:
            (None)

        Returns:
            (int) : The updated count is returned
        """
        self.idx += 1
        return self.idx



"""
Will store all the functions corresponding to Image Assets in lottie
"""

import os
import sys
import struct
import imghdr
import shutil
import settings
sys.path.append("..")


def get_image_size(fname):
    '''
    https://stackoverflow.com/questions/8032642/how-to-obtain-image-size-using-standard-python-class-without-using-external-lib
    Determine the image type of fhandle and return its size.
    from draco

    Args:
        fname (str) : File name

    Returns:
        (int, int) : width and height of image file is returned
        (None)     : If some exception occurs while calculating
    '''
    with open(fname, 'rb') as fhandle:
        head = fhandle.read(24)
        if len(head) != 24:
            return
        if imghdr.what(fname) == 'png':
            check = struct.unpack('>i', head[4:8])[0]
            if check != 0x0d0a1a0a:
                return
            width, height = struct.unpack('>ii', head[16:24])
        elif imghdr.what(fname) == 'gif':
            width, height = struct.unpack('<HH', head[6:10])
        elif imghdr.what(fname) == 'jpeg':
            try:
                fhandle.seek(0) # Read 0xff next
                size = 2
                ftype = 0
                while not 0xc0 <= ftype <= 0xcf:
                    fhandle.seek(size, 1)
                    byte = fhandle.read(1)
                    while ord(byte) == 0xff:
                        byte = fhandle.read(1)
                    ftype = ord(byte)
                    size = struct.unpack('>H', fhandle.read(2))[0] - 2
                # We are at a SOFn block
                fhandle.seek(1, 1)  # Skip `precision' byte.
                height, width = struct.unpack('>HH', fhandle.read(4))
            except Exception: #IGNORE:W0703
                return
        else:
            return
        return width, height


def add_image_asset(lottie, layer):
    """
    Generates the dictionary corresponding to sources/image.json
    Returns: st required in calling function

    Args:
        lottie (dict)                : Lottie layer
        layer  (commong.Layer.Layer) : Synfig layer

    Returns:
        (dict) : Stores address of parameters: "tl", "br", "filename"
    """
    lottie["id"] = "image_" + str(settings.num_images.inc())
    st = {}     # Store the address of children

    st["tl"] = layer.get_param("tl")
    st["br"] = layer.get_param("br")
    st["filename"] = layer.get_param("filename")

    file_path = os.path.join(settings.file_name["fd"], st["filename"][0].text)
    file_path = os.path.abspath(file_path)
    width, height = get_image_size(file_path)
    lottie["w"] = width

    lottie["h"] = height

    images_dir = os.path.join(settings.file_name["fd"], "images")
    images_dir = os.path.abspath(images_dir)
    if not os.path.isdir(images_dir):   # Create images directory if not present
        try:
            os.mkdir(images_dir)
        except OSError:
            print("Creation of the directory %s failed" % images_dir)

    # copy original image to images directory
    src = file_path
    head, tail = os.path.split(file_path)
    new_image_path = os.path.join(images_dir, tail)

    # using shutil to make a copy of the image
    shutil.copy(src, new_image_path)

    # copy meta-data of the file
    shutil.copystat(src, new_image_path)

    lottie["u"] = "images/"
    lottie["p"] = tail
    return st



# pylint: disable=line-too-long
"""
Angle.py
This module will store the Angle class

CHECK:
https://github.com/synfig/synfig/blob/15607089680af560ad031465d31878425af927eb/ETL/ETL/_angle.h
"""

import sys
import math
from common.Vector import Vector
from functools import total_ordering
sys.path.append("..")

ANGLE_EPSILON = 1e-6
PI = 3.1415926535897932384626433832795029

@total_ordering
class Angle:
    """
    Class to keep Synfig format Angle
     
    """
    def __init__(self, v = 0):
        self.v = v

    def __add__(self, other):
        addition = self.v + other.v
        return Angle(addition)

    def __sub__(self, other):
        sub = self.v - other.v
        return Angle(sub)

    def __mul__(self, other):
        if isinstance(other, self.__class__):
            return Angle(self.v * other.v)
        return Angle(self.v * other)

    def __rmul__(self, other):
        return self.__mul__(other)

    def __truediv__(self, other):
        if isinstance(other, self.__class__):
            return Angle(self.v / other.v)
        return Angle(self.v / other)

    def __neg__(self):
        return Angle(-self.v)

    def __lt__(self, rhs):
        return self.v < rhs.v

    def __eq__(self, rhs):
        return math.fabs(self.v - rhs.v) < ANGLE_EPSILON

    def dist(self, rhs):
        return Angle(self.v) - rhs


class RadAngle(Angle):
    """
    Class to keep the angle in Radian format
    """
    def __init__(self, v):
        if isinstance(v, Angle):
            Angle.__init__(self, v.v)
        else:
            Angle.__init__(self, v)

    def get(self):
        return self.v


class DegAngle(Angle):
    """
    Class to keep the angle in Degree format
    """
    def __init__(self, x):
        if isinstance(x, Angle):
            Angle.__init__(self, x.v)
        else:
            v = x * ((PI*2)/360)
            Angle.__init__(self, v)

    def get(self):
        return self.v * 360 / (PI*2)

class SinAngle(Angle):
    """
    Class to keep the angle in Sin format
    """
    def __init__(self, x):
        if isinstance(x, Angle):
            Angle.__init__(self, x.v)
        else:
            v = math.asin(x)
            Angle.__init__(self, v)

    def get(self):
        return math.sin(self.v)

class CosAngle(Angle):
    """
    Class to keep the angle in Cos format
    """
    def __init__(self, x):
        if isinstance(x, Angle):
            Angle.__init__(self, x.v)
        else:
            v = math.acos(x)
            Angle.__init__(self, v)

    def get(self):
        return math.cos(self.v)

class TanAngle(Angle):
    """
    Class to keep the angle in Tan format
    """
    def __init__(self, x):
        if isinstance(x, Angle):
            Angle.__init__(self, x.v)
        elif isinstance(x, Vector):
            v = math.atan2(x[0], x[1])
            Angle.__init__(self, v)
        else:
            v = math.atan(x)
            Angle.__init__(self, v)

    def get(self):
        return math.tan(self.v)



"""
Will store all the functions needed to export the translate layer
"""

import sys
from lxml import etree
import settings
from common.Param import Param
from helpers.transform import gen_helpers_transform
sys.path.append("..")


def gen_layer_translate(lottie, layer):
    """
    Help generate transform properties of translate layer

    Args:
        lottie (dict) : Transform properties in lottie format
        layer  (common.Layer.Layer) : Transform properties in Synfig format

    Returns:
        (None)
    """

    origin = layer.get_param("origin")
    origin.animate("vector")
    pos = origin

    st = "<param name='anchor'><vector><x>0.00</x><y>0.00</y></vector></param>"
    anchor = etree.fromstring(st)
    anchor = Param(anchor, layer)
    anchor.animate("vector")
    anchor.add_offset()

    if settings.INSIDE_PRECOMP:
        pos.add_offset()
    anchor.animate("vector", True)
    pos.animate("vector", True)

    gen_helpers_transform(lottie, pos, anchor)



"""
BlinePoint.py
To store a single Bline point, this class will be used

REFER HERE:
https://github.com/synfig/synfig/blob/15607089680af560ad031465d31878425af927eb/synfig-core/src/synfig/dashitem.cpp
"""

import sys
from common.Vector import Vector
sys.path.append("..")


class BlinePoint:
    """
    Class to store Synfig format Bline points
    TODO: origin_ is not kept as a parameter for Blinepoint because it is not
    yet used. We can set it as a parameter once we know where to use it.
    """
    def __init__(self, vertex_, width_, split_tangent_radius_, split_tangent_angle_, tangent_1, tangent_2, origin_ = 0, boned_vertex_ = False):
        self.vertex_ = vertex_
        self.width_ = width_
        self.origin_ = 0
        self.split_tangent_radius_ = split_tangent_radius_
        self.split_tangent_angle_ = split_tangent_angle_
        self.boned_vertex_ = boned_vertex_
        self.tangent_ = []
        self.tangent_.append(tangent_1)
        self.tangent_.append(tangent_2)
        self.origin_ = origin_
        self.vertex_setup_ = vertex_
        self.update_flags()
        self.update_tangent2()

    def update_flags(self):
        self.split_tangent_both_ = self.split_tangent_radius_ and self.split_tangent_angle_
        self.merge_tangent_both_ = (not self.split_tangent_radius_) and (not self.split_tangent_angle_)

    def update_tangent2(self):
        if self.tangent_[0].mag_squared() != 0:
            self.tangent2_radius_split_ = Vector(self.tangent_[1].mag(), self.tangent_[0].angle())
        else:
            self.tangent2_radius_split_ = self.tangent_[1]
        self.tangent2_angle_split_ = Vector(self.tangent_[0].mag(), self.tangent_[1].angle())

    def get_tangent2(self):
        if self.merge_tangent_both_:
            return self.tangent_[0]
        if self.split_tangent_both_:
            return self.tangent_[1]
        if self.split_tangent_radius_:
            return self.tangent2_radius_split_
        return self.tangent2_angle_split_

    def get_tangent1(self):
        return self.tangent_[0]

    def set_tangent1(self, x):
        self.tangent_[0] = x
        self.update_tangent2()

    def set_tangent2(self, x):
        self.tangent_[1] = x
        self.update_tangent2()

    def get_vertex(self):
        return self.vertex_

    def get_split_tangent_angle(self):
        return self.split_tangent_angle_

    def get_split_tangent_radius(self):
        return self.split_tangent_radius_

    def get_width(self):
        return self.width_

    def set_width(self, x):
        self.width_ = x

    def set_origin(self, x):
        self.origin_ = x

    def get_origin(self):
        return self.origin_

    def get_boned_vertex_flag(self):
        return self.boned_vertex_

    def set_split_tangent_both(self, x=True):
        self.set_split_tangent_radius(x)
        self.set_split_tangent_angle(x)

    def get_split_tangent_both(self):
        return self.split_tangent_both_

    def set_split_tangent_radius(self, x=True):
        self.split_tangent_radius_ = x
        self.update_tangent2()
        self.update_flags()

    def set_split_tangent_angle(self, x=True):
        self.split_tangent_angle_ = x
        self.update_tangent2()
        self.update_flags()

    def set_vertex(self, x):
        self.vertex_ = x
        self.vertex_setup_ = self.vertex_



# pylint: disable=line-too-long
"""
Will store all the functions and modules for generation of outline layer
in Lottie format without variable widths
"""

import sys
import settings
from common.Bline import Bline
from common.misc import is_animated
from common.Count import Count
from properties.value import gen_properties_value
from properties.valueKeyframed import gen_value_Keyframed
from properties.shapePropKeyframe.helper import insert_dict_at, animate_tangents, cubic_to

def gen_bline_outline_constant(lottie, bline_point, layer, transformation, idx):
	"""
	"""
	index = Count()
	lottie["ty"]  = "gr"
	lottie["nm"]  = "Shape "+ str(idx)
	lottie["np"]  = 3
	lottie["cix"] = 2
	lottie["bm"]  = 0
	lottie["ix"]  = idx
	lottie["mn"]  = "ADBE Vector Group " +str(idx)
	lottie["hd"]  = "false"
	lottie["it"]  = []
	lottie["it"].append({})
	lottie["it"].append({})
	lottie["it"].append({})

	bline = Bline(bline_point[0], bline_point)

	#Creating transformation dictionary
	lottie["it"][2]["ty"] = "tr"
	lottie["it"][2]["nm"] = "Transform"
	lottie["it"][2].update(transformation)
	#Creating stroke dictionary
	lottie["it"][1]["ty"] = "st"
	lottie["it"][1]["lc"] = 2
	lottie["it"][1]["lj"] = 1
	lottie["it"][1]["ml"] = 37
	lottie["it"][1]["bm"] = 0
	lottie["it"][1]["nm"] = "Stroke " + str(idx)
	lottie["it"][1]["mn"] = "ADBE Vector Graphic - Stroke " +str(idx)
	lottie["it"][1]["hd"] = "false"
	lottie["it"][1]["c"]  = {}
	lottie["it"][1]["o"]  = {}
	lottie["it"][1]["w"]  = {}

	#Color
	color = layer.get_param("color").get()
	is_animate = is_animated(color[0])
	if is_animate == settings.ANIMATED:
		gen_value_Keyframed(lottie["it"][1]["c"], color[0], index.inc())

	else:
		if is_animate == settings.NOT_ANIMATED:
			val = color[0]
		else:
			val = color[0][0][0]
		red = float(val[0].text)
		green = float(val[1].text)
		blue = float(val[2].text)
		red, green, blue = red ** (1/settings.GAMMA[0]), green ** (1/settings.GAMMA[1]), blue ** (1/ settings.GAMMA[2])
		alpha = float(val[3].text)
		gen_properties_value(lottie["it"][1]["c"],
							[red, green, blue, alpha],
							index.inc(),
							settings.DEFAULT_ANIMATED,
							settings.NO_INFO)

	#Opacity
	opacity = layer.get_param("amount")
	opacity.animate("opacity")
	opacity.fill_path(lottie["it"][1],"o")

	#Constant width
	loop = bline.get_loop()
	width = layer.get_param("width")
	width.scale_convert_link(1)
	width.animate("real")
	width.fill_path(lottie["it"][1],"w")

	#Creating shape dictionary
	lottie["it"][0]["ind"] =  0
	lottie["it"][0]["ty"] = "sh"
	lottie["it"][0]["ix"] = 1
	lottie["it"][0]["ks"] = {}
	lottie["it"][0]["nm"] =  "Path 1"
	lottie["it"][0]["mn"] =  "ADBE Vector Shape - Group"
	lottie["it"][0]["hd"] =  "false"
	lottie["it"][0]["ks"]["a"] = 1 
	lottie["it"][0]["ks"]["ix"] = lottie["it"][0]["ix"] + 1 
	lottie["it"][0]["ks"]["k"] = []

	window = {}
	window["first"] = sys.maxsize
	window["last"] = -1

	for entry in bline.get_entry_list():
		pos = entry["point"]
		width = entry["width"]
		t1 = entry["t1"]
		t2 = entry["t2"]
		split_r = entry["split_radius"]
		split_a = entry["split_angle"]

		pos.update_frame_window(window)
		# Empty the pos and fill in the new animated pos
		pos.animate("vector")

		width.update_frame_window(window)
		width.animate("real")

		split_r.update_frame_window(window)
		split_r.animate_without_path("bool")

		split_a.update_frame_window(window)
		split_a.animate_without_path("bool")

		animate_tangents(t1, window)
		animate_tangents(t2, window)


	outer_width = layer.get_param("width")
	sharp_cusps = layer.get_param("sharp_cusps")
	expand = layer.get_param("expand")
	r_tip0 = layer.get_param("round_tip[0]")
	r_tip1 = layer.get_param("round_tip[1]")
	homo_width = layer.get_param("homogeneous_width")
	origin = layer.get_param("origin")

	# Animating the origin
	origin.update_frame_window(window)
	origin.animate("vector")

	# Animating the outer width
	outer_width.update_frame_window(window)
	outer_width.animate("real")

	# Animating the sharp_cusps
	sharp_cusps.update_frame_window(window)
	sharp_cusps.animate_without_path("bool")

	# Animating the expand param
	expand.update_frame_window(window)
	expand.animate("real")

	# Animating the round tip 0
	r_tip0.update_frame_window(window)
	r_tip0.animate_without_path("bool")

	# Animating the round tip 1
	r_tip1.update_frame_window(window)
	r_tip1.animate_without_path("bool")

	# Animating the homogeneous width
	homo_width.update_frame_window(window)
	homo_width.animate_without_path("bool")
	# Minimizing the window size
	if window["first"] == sys.maxsize and window["last"] == -1:
		window["first"] = window["last"] = 0
	
	frames = list(set(settings.WAYPOINTS_LIST))
	length = bline.get_len()

	flag = False
	if loop:
		flag = True

	for fr in frames:
		if fr!=1:
			st_val = insert_dict_at(lottie["it"][0]["ks"]["k"], -1, fr, flag,True)
			cur_origin = origin.get_value(fr)
			for point in range(0,length):
				pos_ret, width, t1, t2, split_r_val, split_a_val = get_outline_param_at_frame(bline[point],fr)
				cubic_to(pos_ret,t1,t2,st_val,cur_origin,False,True)


def get_outline_param_at_frame(entry, fr):
    """
    Given a entry and frame, returns the parameters of the outline layer at
    that frame

    Args:
        entry (dict) : Vertex of outline layer in Synfig format
        fr        (int)                 : frame number

    Returns:
        (common.Vector.Vector) : position of the vertex
        (float)       : width of the vertex
        (common.Vector.Vector) : Tangent 1 of the vertex
        (common.Vector.Vector) : Tangent 2 of the vertex
        (bool)        : True if radius split is ticked at this frame
        (bool)        : True if tangent split is ticked at this frame
    """
    pos = entry["point"].get_value(fr)
    # Convert pos back to Synfig coordinates
    pos = to_Synfig_axis(pos, "vector")
    pos_ret = Vector(pos[0], pos[1])

    width = entry["width"].get_value(fr)
    width = to_Synfig_axis(width, "real")
    t1 = entry["t1"]
    t2 = entry["t2"]
    split_r = entry["split_radius"]
    split_a = entry["split_angle"]


    t1, t2 = get_tangent_at_frame(t1, t2, split_r, split_a, fr)
    # Convert to Synfig units
    t1 /= settings.PIX_PER_UNIT
    t2 /= settings.PIX_PER_UNIT

    split_r_val = split_r.get_value(fr)
    split_a_val = split_a.get_value(fr)

    return pos_ret, width, t1, t2, split_r_val, split_a_val


# pylint: disable=line-too-long
"""
Will store all the functions and modules for generation of circle layer
in Lottie format
"""

import sys
import math
from common.Matrix2 import Matrix2
from common.Vector import Vector
from synfig.animation import to_Synfig_axis
from properties.shapePropKeyframe.helper import add, insert_dict_at, quadratic_to_cubic
sys.path.append("../../")


def gen_list_circle(lottie, layer):
    """
    Generates a shape layer corresponding to circle layer by manipulating the
    origin and radius of the circle

    Args:
        lottie (dict) : Lottie format circle layer will be stored in this
        layer  (common.Layer.Layer) : Synfig format circle layer

    Returns:
        (None)
    """
    ################### SECTION 1 #########################
    # Inserting waypoints if not animated and finding the first and last frame
    window = {}
    window["first"] = sys.maxsize
    window["last"] = -1

    origin = layer.get_param("origin")
    radius = layer.get_param("radius")

    # Animating the origin
    origin.update_frame_window(window)
    origin.animate("vector")

    radius.update_frame_window(window)
    radius.animate("real")

    # Minimizing the window size
    if window["first"] == sys.maxsize and window["last"] == -1:
        window["first"] = window["last"] = 0
    ################# END OF SECTION 1 ###################

    ################ SECTION 2 ###########################
    # Generating values for all the frames in the window

    fr = window["first"]
    while fr <= window["last"]:
        st_val, en_val = insert_dict_at(lottie, -1, fr, False)  # This loop needs to be considered somewhere down

        synfig_circle(st_val, origin, radius, fr)
        synfig_circle(en_val, origin, radius, fr + 1)

        fr += 1
    # Setting the final time
    lottie.append({})
    lottie[-1]["t"] = fr


def synfig_circle(st_val, origin_param, radius_param, fr):
    """
    Calculates the points for the circle layer as in Synfig:
    https://github.com/synfig/synfig/blob/678cc3a7b1208fcca18c8b54a29a20576c499927/synfig-core/src/modules/mod_geometry/circle.cpp

    Args:
        st_val (dict) : Lottie format circle is stored in this
        origin (common.Param.Param) : Lottie format origin of circle
        radius (common.Param.Param) : Lottie format radius of circle
        fr (int) : Frame number

    Returns:
        (None)
    """

    num_splines = 8
    angle = 180.0 / num_splines
    angle *= ((math.pi * 2) / 360)
    k = 1.0 / math.cos(angle)

    radius = abs(to_Synfig_axis(radius_param.get_value(fr), "real"))
    origin = origin_param.get_value(fr)

    matrix = Matrix2()
    matrix.set_rotate(angle)

    p0, p1, p2 = Vector(), Vector(), Vector(radius, 0)

    # Setup chunk list
    chunk_list = []
    chunk_list.append([p2, Vector(), Vector()])
    i = 0
    while i < num_splines:
        p0 = p2
        p1 = matrix.get_transformed(p0)
        p2 = matrix.get_transformed(p1)
        cp1, cp2 = quadratic_to_cubic(p2, k*p1, p0)
        cur_tan = p2 - cp1
        chunk_list[-1][2] = cp2 - p0
        chunk_list.append([p2, cur_tan, Vector()])
        i += 1

    add(chunk_list, st_val, origin)
